/*
** clock.c - Test for programmable clock generators; user can specify a
**           clock generator to program, and the application (approximately)
**           measures the actual frequency.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**   o ADM-XRC-5T2-ADV-CC1
**   o ADM-PCIE-6S1
**
** (C) Copyright 2008-2014 Alpha Data
**
** Cycle counters map to clock generators on a per-model basis as
** follows:
**
** Model               Cycle counter index      Clock generator index
** -------------------------------------------------------------------
**
** ADM-XRC             LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   MCLK (1)
**
** ADM-XRC-P           LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   MCLK (1)
**
** ADM-XRC-II-Lite     LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   MCLK (1)
**
** ADM-XRC-II          LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   MCLK (1)
**
** ADM-XPL             LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   LCLK * 2 (0)
**                     REFCLK pin    (3)   =>   N/A (fixed @ 125 MHz)
**
** ADM-XP              LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   LCLK * 2 (0)
**                     REFCLK pin    (3)   =>   N/A (fixed @ 125 MHz)
**
** ADP-XPI             LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   LCLK * 2 (0)
**                     REFCLK pin    (3)   =>   N/A (fixed @ 125 MHz)
**
** ADP-WRC-II          LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   MCLK (1)
**
** ADP-DRC-II          LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   MCLK (1)
**                     DDRCLK pin    (3)   =>   N/A (fixed @ 66 MHz)
**
** ADM-XRC-4LX         LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   MCLK (1)
**                     REFCLK pin    (3)   =>   N/A (fixed @ 200 MHz)
**
** ADM-XRC-4SX         LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   MCLK (1)
**                     REFCLK pin    (3)   =>   N/A (fixed @ 200 MHz)
**
** ADM-XRC-4FX         LCLK pin      (1)   =>   LCLK (0)
**                     MCLKA pin     (2)   =>   MCLK (1)
**                     MCLKB pin     (3)   =>   MCLK (1)
**                     REFCLK200 pin (4)   =>   N/A (fixed @ 200 MHz)
**                     XRMCLK pin    (5)   =>   N/A (depends on I/O pin)
**                     XRMMGTREF pin (6)   =>   N/A (depends on I/O pin)
**                     PCIE250A pin  (7)   =>   N/A (fixed @ 250 MHz)
**
** ADPE-XRC-4FX        LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   MCLK (1)
**                     REFCLK200 pin (3)   =>   N/A (fixed @ 200 MHz)
**
** ADM-XRC-5LX         LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   MCLK (1)
**                     REFCLK200 pin (3)   =>   N/A (fixed @ 200 MHz)
**                     XRMCLK pin    (4)   =>   N/A (depends on I/O pin)
**
** ADM-XRC-5T1         LCLK pin      (1)   =>   LCLK (0)
**                     MCLKA pin     (2)   =>   MCLK (1)
**                     MCLKB pin     (3)   =>   MCLK (1)
**                     REFCLK200 pin (4)   =>   N/A (fixed @ 200 MHz)
**                     XRMCLK pin    (5)   =>   N/A (depends on I/O pin)
**                     XRMMGTREF pin (6)   =>   N/A (depends on I/O pin)
**                     PCIE100A pin  (7)   =>   N/A (fixed @ 100 MHz)
**
** ADM-XRC-5T2 or      LCLK pin      (1)   =>   LCLK (0)
** ADM-XRC-5T2-ADV or  MCLKA pin     (2)   =>   MCLK (1)
** ADM-XRC-5T2-ADV6 or MCLKB pin     (3)   =>   MCLK (1)
** ADM-XRC-5T2-ADV-CC1 REFCLK200 pin (4)   =>   N/A (fixed @ 200 MHz)
**                     XRMCLK pin    (5)   =>   N/A (depends on I/O pin)
**                     XRMMGTREF pin (6)   =>   N/A (depends on I/O pin)
**                     PCIE100A pin  (7)   =>   N/A (fixed @ 100 MHz)
**
** ADM-XRC-5TZ         LCLK pin      (1)   =>   LCLK (0)
**                     MCLKA pin     (2)   =>   MCLK (1)
**                     MCLKB pin     (3)   =>   MCLK (1)
**                     REFCLK200 pin (4)   =>   N/A (fixed @ 200 MHz)
**                     XRMCLK pin    (5)   =>   N/A (depends on I/O pin)
**                     XRMMGTREF pin (6)   =>   N/A (depends on I/O pin)
**                     PCIE100A pin  (7)   =>   N/A (fixed @ 100 MHz)
**
** ADM-XRC-5T-DA1      LCLK pin      (1)   =>   LCLK (0)
**                     MCLKA pin     (2)   =>   MCLK (1)
**                     MCLKB pin     (3)   =>   MCLK (1)
**                     REFCLK200 pin (4)   =>   N/A (fixed @ 200 MHz)
**                     PCIE100A pin  (5)   =>   N/A (fixed @ 100 MHz)
**
** ADM-PCIE-6S1        LCLK pin      (1)   =>   LCLK (0)
**                     MCLK pin      (2)   =>   MCLK (1)
**                     REFCLK200 pin (3)   =>   REFCLK200 (2)
**
** Note that this application does not attempt to compensate for any
** innacuracy in reported measurements resulting from:
**
**   (a) limited system timer resolution, and
**
**   (b) preemptive scheduling resulting delaying the reads of a cycle
**       counters due to the system being busy executing other processes
**       or threads.
**
** This application is intended to provide a ball-park measurement of
** clock frequency, rather than a precise measurement.
*/

#include <stdio.h>
#include <stdlib.h>

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _tmain main
# define _T(c) c
# define _tprintf printf
# define _gettchar getchar

typedef unsigned long ULONG;

#endif

#include <admxrc2.h>

#include <sdk_common.h>

#include <platform_endian.h>
#include <platform_time.h>

#define READ_COUNT_REG_INDEX        (0)
#define READY_REG_INDEX             (1)
#define COUNT_VAL_REG_INDEX(n)      (32 + (n))

/*
** Returns TRUE if ready flag for the specified clock input is high
** within 2 milliseconds (give or take the system timer resolution.
*/
static bool_t
waitReady(
  volatile uint32_t* pFpgaSpace,
  unsigned int clockInput)
{
  uint32_t val32;
  admxrc2_time_t start, expire, now;
  unsigned int clockInputIndex = clockInput - 1;

  start = timeGet();
  expire = timeAdd(start, doubleToTime(2.0));
  while (1) {
    val32 = pFpgaSpace[READY_REG_INDEX];
    if (val32 & (0x1U << clockInputIndex)) {
      break;
    }
    now = timeGet();
    if (timeAfter(now, expire)) {
      return FALSE;
    }
  }
  return TRUE;
}

/*
** Performs a read of a cycle counter.
**
** Returns TRUE if timeout NOT detected, i.e. the clock for that cycle
** counter appears to be running.
*/
static bool_t
readCounter(
  volatile uint32_t* pFpgaSpace,
  unsigned int clockInput,
  uint32_t* pCountValue)
{
  unsigned int clockInputIndex = clockInput - 1;

  if (!waitReady(pFpgaSpace, clockInput)) {
    return FALSE;
  }
  pFpgaSpace[READ_COUNT_REG_INDEX] = 0x1U << clockInputIndex;
  if (!waitReady(pFpgaSpace, clockInput)) {
    return FALSE;
  }
  *pCountValue = pFpgaSpace[COUNT_VAL_REG_INDEX(clockInputIndex)];
  return TRUE;
}

static const TCHAR*
getClockInfo(
  ADMXRC2_CARD_INFO cardInfo,
  unsigned int clockInput,
  bool_t* bProgrammable,
  unsigned int* clockGen)
{
  if (clockInput == 1) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMPCIE6S1:
      *bProgrammable = TRUE;
      *clockGen = 0;
      return _T("LCLK");
    default:
      *bProgrammable = TRUE;
      *clockGen = 0;
      return _T("LCLK");
    }
  } else if (clockInput == 2) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC:
    case ADMXRC2_BOARD_ADMXRC_P:
    case ADMXRC2_BOARD_ADMXRC2_LITE:
    case ADMXRC2_BOARD_ADMXRC2:
    case ADMXRC2_BOARD_ADPWRC2:
    case ADMXRC2_BOARD_ADPDRC2:
    case ADMXRC2_BOARD_ADMXRC4LS:
    case ADMXRC2_BOARD_ADMXRC4LX:
    case ADMXRC2_BOARD_ADMXRC4SX:
    case ADMXRC2_BOARD_ADPEXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5LX:
    case ADMXRC2_BOARD_ADMPCIE6S1:
      *bProgrammable = TRUE;
      *clockGen = 1;
      return _T("MCLK");
    case ADMXRC2_BOARD_ADMXPL:
    case ADMXRC2_BOARD_ADMXP:
    case ADMXRC2_BOARD_ADPXPI:
      *bProgrammable = TRUE;
      *clockGen = 1;
      return _T("MCLK");
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
    case ADMXRC2_BOARD_ADMXRC5TDA1:
    case ADMXRC2_BOARD_ADMXRC5T2ADV6:
    case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
      *bProgrammable = TRUE;
      *clockGen = 1;
      return _T("MCLKA");
    default:
      *bProgrammable = FALSE;
      return _T("unknown");
    }
  } else if (clockInput == 3) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXPL:
    case ADMXRC2_BOARD_ADMXP:
    case ADMXRC2_BOARD_ADPXPI:
    case ADMXRC2_BOARD_ADMXRC4LX:
    case ADMXRC2_BOARD_ADMXRC4SX:
    case ADMXRC2_BOARD_ADMXRC5LX:
    case ADMXRC2_BOARD_ADPEXRC4FX:
      *bProgrammable = FALSE;
      return _T("REFCLK");
    case ADMXRC2_BOARD_ADPDRC2:
      *bProgrammable = FALSE;
      return _T("DDRCLK");
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
    case ADMXRC2_BOARD_ADMXRC5TDA1:
    case ADMXRC2_BOARD_ADMXRC5T2ADV6:
    case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
      *bProgrammable = TRUE;
      *clockGen = 1;
      return _T("MCLKB");
    case ADMXRC2_BOARD_ADMPCIE6S1:
      *bProgrammable = TRUE;
      *clockGen = 2;
      return _T("REFCLK200");
    default:
      *bProgrammable = FALSE;
      return _T("unknown");
    }
  } else if (clockInput == 4) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
    case ADMXRC2_BOARD_ADMXRC5TDA1:
    case ADMXRC2_BOARD_ADMXRC5T2ADV6:
    case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
      *bProgrammable = FALSE;
      return _T("REFCLK");
    case ADMXRC2_BOARD_ADMXRC5LX:
      *bProgrammable = FALSE;
      return _T("XRM_CLKIN");
    default:
      *bProgrammable = FALSE;
      return _T("unknown");
    }
  } else if (clockInput == 5) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
    case ADMXRC2_BOARD_ADMXRC5T2ADV6:
    case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
      *bProgrammable = FALSE;
      return _T("XRM_CLKIN");
    case ADMXRC2_BOARD_ADMXRC5TDA1:
      *bProgrammable = FALSE;
      return _T("PCIE100A");
    default:
      *bProgrammable = FALSE;
      return _T("unknown");
    }
  } else if (clockInput == 6) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
    case ADMXRC2_BOARD_ADMXRC5T2ADV6:
    case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
      *bProgrammable = FALSE;
      return _T("XRM_MGTREF");
    default:
      *bProgrammable = FALSE;
      return _T("unknown");
    }
  } else if (clockInput == 7) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC4FX:
      *bProgrammable = FALSE;
      return _T("PCIE250A");
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
    case ADMXRC2_BOARD_ADMXRC5T2ADV6:
    case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
      *bProgrammable = FALSE;
      return _T("PCIE100A");
    default:
      *bProgrammable = FALSE;
      return _T("unknown");
    }
  } else {
    *bProgrammable = FALSE;
    return _T("unknown");
  }
}

int
_tmain(
  int argc,
  TCHAR *argv[])
{
  int ret = 0;
  ADMXRC2_STATUS status;
  ADMXRC2_HANDLE hCard = ADMXRC2_HANDLE_INVALID_VALUE;
  ADMXRC2_CARD_INFO cardInfo;
  ADMXRC2_SPACE_INFO spInfo;
  volatile uint32_t* pFpgaSpace;
  const TCHAR* filename;
  Option options[3];
  OptionType paramType[2];
  TCHAR* paramInfo[2];
  TCHAR* paramMeaning[2];
  Arguments arguments;
  unsigned int clockInput;
  unsigned int numClockInput;
  uint32_t value1, value2, delta;
  bool_t bSuccess;
  bool_t bMeasure;
  double lclkMax;

  /* Set up expected options */
  options[0].key = _T("card");
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("ID of card to use");
  options[1].key = _T("index");
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("index of card to use");
  options[2].key = _T("measure");
  options[2].type = Option_boolean;
  options[2].def.booleanVal = TRUE;
  options[2].meaning = _T("TRUE to measure frequency");
  arguments.nOption = 3;
  arguments.option = options;

  paramType[0] = Option_uint;
  paramInfo[0] = _T("clockInput");
  paramMeaning[0] = _T("Index of clock generator to test");
  paramType[1] = Option_double;
  paramInfo[1] = _T("frequency");
  paramMeaning[1] = _T("Requested frequency (MHz)");
  arguments.nParamType = 2;
  arguments.minNParam = 0;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;

  /* Parse command line */
  parse_command_line(_T("CLOCK test"), _T("clock"), argc, argv, &arguments);

  if (options[0].specified && options[1].specified) {
    _tprintf(_T("Do not specify both -card and -index\n"));
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Failed to open card with ID %lu: %s\n"),
        (unsigned long)cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Failed to open card with index %lu: %s\n"),
        (unsigned long)index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  bMeasure = options[2].value.booleanVal;

  status = ADMXRC2_GetCardInfo(hCard, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("Failed to get card info: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
    numClockInput = 2; /* LCLK, MCLK */
    lclkMax = 40.0e6;
    break;

  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADPWRC2:
    numClockInput = 2; /* LCLK, MCLK */
    lclkMax = 66.67e6;
    break;

  case ADMXRC2_BOARD_ADPDRC2:
    numClockInput = 3; /* LCLK, MCLK, DDRCLK */
    lclkMax = 66.67e6;
    break;

  case ADMXRC2_BOARD_ADMXPL:
  case ADMXRC2_BOARD_ADMXP:
  case ADMXRC2_BOARD_ADPXPI:
    numClockInput = 3; /* LCLK, MCLK, REFCLK */
    lclkMax = 80.0e6;
    break;

  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
  case ADMXRC2_BOARD_ADPEXRC4FX:
    numClockInput = 3; /* LCLK, MCLK, REFCLK */
    lclkMax = 50.0e6;
    break;

  case ADMXRC2_BOARD_ADMXRC4FX:
    numClockInput = 7; /* LCLK, MCLKA, MCLKB, REFCLK, XRMCLK, XRMMGTREF, PCIE250A */
    lclkMax = 50.0e6;
    break;

  case ADMXRC2_BOARD_ADMXRC5LX:
    numClockInput = 4; /* LCLK, MCLK, REFCLK, XRMCLK */
    lclkMax = 50.0e6;
    break;

  case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMXRC5TZ:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
  case ADMXRC2_BOARD_ADMXRC5T2ADV6:
  case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
    numClockInput = 7; /* LCLK, MCLKA, MCLKB, REFCLK, XRMCLK, XRMMGTREF, PCIE100A */
    lclkMax = 50.0e6;
    break;

  case ADMXRC2_BOARD_ADMXRC5TDA1:
    numClockInput = 5; /* LCLK, MCLKA, MCLKB, REFCLK, PCIE100A */
    lclkMax = 50.0e6;
    break;

  case ADMXRC2_BOARD_ADMPCIE6S1:
    numClockInput = 3; /* LCLK, MCLK, REFCLK200 */
    lclkMax = 66.0e6;
    break;

  default:
    _tprintf("*** This example must be run on one of the following cards:\n\n");
    _tprintf("\tADM-XRC\n");
    _tprintf("\tADM-XRC-P\n");
    _tprintf("\tADM-XRC-II-Lite\n");
    _tprintf("\tADM-XRC-II\n");
    _tprintf("\tADM-XPL\n");
    _tprintf("\tADM-XP\n");
    _tprintf("\tADP-WRC-II\n");
    _tprintf("\tADP-DRC-II\n");
    _tprintf("\tADP-XPI\n");
    _tprintf("\tADM-XRC-4LX\n");
    _tprintf("\tADM-XRC-4SX\n");
    _tprintf("\tADM-XRC-4FX / ADM-XMC-4FX\n");
    _tprintf("\tADPE-XRC-4FX\n");
    _tprintf("\tADM-XRC-5LX\n");
    _tprintf("\tADM-XRC-5T1\n");
    _tprintf("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6\n");
    _tprintf("\tADM-XRC-5TZ\n");
    _tprintf("\tADM-XRC-5T-DA1\n");
    _tprintf("\tADM-XRC-5T2-ADV-CC1\n");
    _tprintf("\tADM-PCIE-6S1\n");
    ret = -1;
    goto done;
  }

  /* Get the address of FPGA space */
  status = ADMXRC2_GetSpaceInfo(hCard, 0, &spInfo);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf("Failed to get space 0 info: %s\n",
      ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  pFpgaSpace = (volatile uint32_t*)spInfo.VirtualBase;
  if (pFpgaSpace == NULL) {
    _tprintf("*** Failed to retrieve pointer to FPGA space\n");
    return -1;
  }

  if (arguments.nParam >= 1) {
    const TCHAR* pClockName;
    unsigned int clockGen;
    bool_t bProgrammable;

    clockInput = (unsigned int)arguments.param[0].uintVal;

    if (clockInput < 1 || clockInput > 32) {
      _tprintf("*** Clock index of %lu is invalid; must be in the range 1..32\n",
        (unsigned long)clockInput);
      ret = -1;
      goto done;
    }
    if (clockInput > numClockInput) {
      _tprintf("*** Clock index of %lu is invalid; maximum value for this card is %lu.\n",
        (unsigned long)clockInput, (unsigned long)numClockInput);
      ret = -1;
      goto done;
    }

    pClockName = getClockInfo(cardInfo, clockInput, &bProgrammable, &clockGen);

    if (bMeasure) {
      filename = get_fpga_filename(&cardInfo, "clock", &arguments);
      status = ADMXRC2_ConfigureFromFile(hCard, filename);
      if (status != ADMXRC2_SUCCESS) {
        _tprintf ("Failed to load the bitstream '%s': %s\n",
          filename, ADMXRC2_GetStatusString(status));
        ret = -1;
        goto done;
      }
    }

    if (arguments.nParam >= 2) {
      double reqFreq, progFreq;

      if (bProgrammable) {

        /* Set specified clock generator to requested frequency */
        reqFreq = arguments.param[1].doubleVal * 1000000.0;
        if (clockInput == 1) {
          if (reqFreq > lclkMax) {
            _tprintf("*** Maximum LCLK frequency for this model is %.2f MHz\n",
              lclkMax / 1.0e6);
            ret = -1;
            goto done;
          }
        }

        _tprintf("Programming clock generator %lu for %.2f MHz...\n",
          (unsigned long)clockGen, reqFreq / 1.0e6);
        status = ADMXRC2_SetClockRate(hCard, clockGen, reqFreq, &progFreq);
        if (status != ADMXRC2_SUCCESS) {
          _tprintf("*** Failed to program clock generator %lu for %.2f MHz: %s\n",
            (unsigned long)clockGen,
            reqFreq / 1.0e6,
            ADMXRC2_GetStatusString(status));
          ret = -1;
          goto done;
        }
        _tprintf("Actual programmed frequency = %.2f Hz\n", progFreq);
      } else {
        _tprintf("*** Clock input %lu (%s) is not programmable\n",
          (unsigned long)clockInput, pClockName);
        ret = -1;
        goto done;
      }
    }       

    if (bMeasure) {
      _tprintf("Measuring frequency of clock input %lu (%s)...\n",
        (unsigned long)clockInput, pClockName);

      /* Now read the specified cycle counter twice 1 second apart */
      sleepMs(1);
      bSuccess = readCounter(pFpgaSpace, clockInput, &value1);
      if (bSuccess) {
        sleepMs(1000);
        bSuccess = readCounter(pFpgaSpace, clockInput, &value2);
        if (bSuccess) {
          delta = value2 - value1;
          _tprintf("Initial counter value = %lu\n", (unsigned long)value1);
          _tprintf("Final counter value   = %lu, delta = %lu\n",
            (unsigned long)value2, (unsigned long)delta);
        }
      }
      if (!bSuccess) {
        _tprintf("Clock input %lu (%s) appears to be stopped.\n",
          (unsigned long)clockInput, pClockName);
      }
    }
  } else {
    unsigned int i;
    const TCHAR* pClockName;
    bool_t bProgrammable;
    unsigned int clockGen;

    _tprintf("Clock pins available on specified card:\n\n");
    _tprintf("Clock input      Programmable   Clock generator index\n\n");
    for (i = 1; i <= numClockInput; i++) {
      pClockName = getClockInfo(cardInfo, i, &bProgrammable, &clockGen);
      if (bProgrammable) {
        _tprintf("%-2lu %-12s  %-12s   %-lu\n",
          (unsigned long)i, pClockName, bProgrammable ? "Yes" : "No", (unsigned long)clockGen);
      } else {
        _tprintf("%-2lu %-12s  %-12s   N/A\n",
          (unsigned long)i, pClockName, bProgrammable ? "Yes" : "No");
      }
    }
  }

done:
  if (hCard != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(hCard);
  }

  return ret;
}
