/*
** dma.c - example demonstrating multithreaded DMA transfers
**         to and from FPGA, using demand mode DMA.
**
** Data is transferred using demand mode DMA to the FPGA, stored
** in a FIFO and transferred out again also using demand mode DMA.
** Also demonstrated is configuring the FPGA using DMA.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**   o ADM-XRC-5T2-ADV-CC1
**   o ADM-PCIE-6S1
**
** (C) Copyright 2000-2014 Alpha Data
**
**
** NOTES ON OPERATION
**
** This application is designed to show that multiple threads can
** use the same DMA channel without error and without affecting
** data integrity.
**
** 1. Threading
**
** A sender DMA thread is started up. This thread transfers data into the
** FIFO in the FPGA.
**
** A receiver DMA thread is started up, which transfers data out of the FIFO
** in the FPGA and checks that it is correct.
**
** 2. DMA mode bits
**
** The DMA transfers are performed with the DEMAND mode bit set.
** Thus, the DMA transfers are regulated by the ability of the FPGA
** to accept or supply data.
**
** Other mode bits used are:
**
**    - FIXEDLOCAL:  the FIFO input and output ports are 32-bit wide
**                   registers on the local buss. Thus a DMA transfer
**                   must not increment the local bus address.
**    - BURSTENABLE: to achieve good throughput, it is necessary to
**                   enable burst transfers on the local bus.
**    - USEBTERM:    the FPGA can terminate a local bus burst by
**                   asserting BTERM.
**    - USEREADY:    the FPGA uses LREADYI# to control when data
**                   transfer can occur within a local bus burst.
**
** 3. DMA driven FPGA configuration
**
** Change the DMA_CONFIGURATION macro from 1 to 0 in order to load
** the FPGA using programmed I/O rather then DMA.
**
** 4. Data integrity checking
**
** By default, this program checks that the data received by the sucker
** thread is correct.
*/

#include <stdio.h>
#include <stdlib.h>

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

#include <stdint.h>
#include <semaphore.h>
#include <pthread.h>

/* Linux or VxWorks */
# define _tmain main
# define _T(c) c
# define _tprintf printf
# define _gettchar getchar

typedef unsigned long ULONG;

#endif

#include <admxrc2.h>

#include <sdk_common.h>

#include <platform_endian.h>
#include <platform_time.h>

static ADMXRC2_HANDLE hCard = ADMXRC2_HANDLE_INVALID_VALUE;
static ULONG mode;
static bool_t bCheckData = TRUE;
static ULONG bufSize;

#if defined(_WIN32)
static HANDLE readySem = NULL;
#else
static sem_t readySem;
bool_t bReadySemInited = FALSE;
#endif

static bool_t bTerminate;
static bool_t bTerminateRecv;

static unsigned int recvChan, sendChan;
#if defined(_WIN32)
static HANDLE hRecvThread, hSendThread;
static DWORD recvThreadId, sendThreadId;
#else
static pthread_t pRecvThread, pSendThread;
#endif
static double received, sent;

static ADMXRC2_DMADESC hSendBuf;
static bool_t bSendBufValid = FALSE;
static uint8_t* pSendBuf = NULL;
static ADMXRC2_DMADESC hRecvBuf;
static bool_t bRecvBufValid = FALSE;
static uint8_t* pRecvBuf = NULL;

static volatile uint32_t* pFpgaSpace;

#if defined(_WIN32)
static DWORD WINAPI
#else
static void*
#endif
sendThreadFunc(
  void* pArg)
{
  ULONG m = bufSize >> 2;
  ADMXRC2_STATUS status;
  ULONG j;
#if defined(_WIN32)
  HANDLE hEvent;
#else
  void* hEvent = NULL;
#endif
  uint32_t val32 = 0x0U;
  uint32_t* pBuf = (uint32_t*)pSendBuf;

#if defined(_WIN32)
  hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
#endif

  /* Set up the data */
  val32 = (uint32_t)0;
  val32 |= val32 << 8;
  val32 |= val32 << 16;
  for (j = 0; j < m; j++) {
    pBuf[j] = val32++;
  }

  while (!bTerminate) {
#if defined(_WIN32)
    ReleaseSemaphore(readySem, 1, NULL);
#else
    sem_post(&readySem);
#endif

    /*
    ** Program the PCI-to-local transfer counter.
    */
    pFpgaSpace[0] = cpu_to_le32(bufSize);
    pFpgaSpace[0];

    /*
    ** Perform the DMA transfer.
    */
    status = ADMXRC2_DoDMA(
      hCard,
      hSendBuf,
      0,
      bufSize,
      0x40,
      ADMXRC2_PCITOLOCAL,
      sendChan,
      mode,
      0,
      NULL,
      hEvent);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Send thread: failed to do DMA: %s\n"), ADMXRC2_GetStatusString(status));
      break;
    }

    sent += bufSize;
  }

#if defined(_WIN32)
  CloseHandle(hEvent);
#endif

  return 0;
}

#if defined(_WIN32)
static DWORD WINAPI
#else
static void*
#endif
recvThreadFunc(
  void* pArg)
{
  const ULONG maxError = 20;
  ULONG m = bufSize >> 2;
  ADMXRC2_STATUS status;
#if defined(_WIN32)
  HANDLE hEvent;
#else
  void* hEvent = NULL;
#endif
  uint32_t* pBuf = (uint32_t*)pRecvBuf;
  ULONG n = 0, nerr;

#if defined(_WIN32)
  hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
#endif

  while (1) {
    ULONG exp, act;
    ULONG j;

#if defined(_WIN32)
    if (WaitForSingleObject(readySem, INFINITE) != WAIT_OBJECT_0) {
      _tprintf(_T("Receive thread: failed to wait on semaphore\n"));
      break;
    }
#else
    sem_wait(&readySem);
#endif

    if (bTerminateRecv) {
      break;
    }

    /*
    ** Program the local-to-PCI transfer counter.
    */
    pFpgaSpace[1] = cpu_to_le32(bufSize);
    pFpgaSpace[1];

    /*
    ** Perform the DMA transfer.
    */
    status = ADMXRC2_DoDMA(
      hCard,
      hRecvBuf,
      0,
      bufSize,
      0x80,
      ADMXRC2_LOCALTOPCI,
      recvChan,
      mode,
      0,
      NULL,
      hEvent);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Receive thread: failed to do DMA: %s\n"), ADMXRC2_GetStatusString(status));
      break;
    }

    if (bCheckData) {
      nerr = 0;
      exp = pBuf[0];
      for (j = 1; j < m; j++) {
        exp++;
        act = pBuf[j];
        if (act != exp) {
          if (nerr < maxError) {
            _tprintf(_T("Receiver thread: error at block %lu index %lu, exp/act=0x%08lX/0x%08lX\n"),
              (unsigned long)n, (unsigned long)j,
              (unsigned long)exp, (unsigned long)act);
          } else if (nerr == maxError) {
            _tprintf(_T("Receiver thread: more than %lu error(s)\n"), maxError);
          }
          nerr++;
        }
      }
    }

    received += bufSize;
    n++;
  }

#if defined(_WIN32)
  CloseHandle(hEvent);
#endif

  return 0;
}

int
_tmain(
  int argc,
  TCHAR* argv[])
{
  int ret = 0;
  ADMXRC2_STATUS status;
  ADMXRC2_CARD_INFO cardInfo;
  ADMXRC2_SPACE_INFO spaceInfo;
  admxrc2_time_t start;
#if defined(_WIN32)
  HANDLE hEvent = NULL;
#else
  void* hEvent = NULL;
#endif
  const TCHAR* filename;
  Arguments arguments;
  Option options[11];
  float clockFreq;
  bool_t bUse64bit, bDmaConfiguration, bCheckData, bInteractive;
  unsigned int secondsToRun;

  options[0].key = _T("card");
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("ID of card to use");
  options[1].key = _T("index");
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("index of card to use");
  options[2].key = _T("size");
  options[2].type = Option_uint;
  options[2].def.uintVal = 64*1024;
  options[2].meaning = _T("send/receive buffer size (bytes)");
  options[3].key = _T("lclk");
  options[3].type = Option_float;
  options[3].def.floatVal = 33.0;
  options[3].meaning = _T("local bus clock frequency (MHz)");
  options[4].key = _T("64");
  options[4].type = Option_boolean;
  options[4].def.booleanVal = FALSE;
  options[4].meaning = _T("TRUE => run local bus in 64 bit mode");
  options[5].key = _T("txchan");
  options[5].type = Option_uint;
  options[5].def.uintVal = 0;
  options[5].meaning = _T("DMA channel to use for sender thread");
  options[6].key = _T("rxchan");
  options[6].type = Option_uint;
  options[6].def.uintVal = 1;
  options[6].meaning = _T("DMA channel to use for receiver thread");
  options[7].key = _T("dmaconfig");
  options[7].type = Option_boolean;
  options[7].def.booleanVal = TRUE;
  options[7].meaning = _T("Use DMA to configure the FPGA");
  options[8].key = _T("check");
  options[8].type = Option_boolean;
  options[8].def.booleanVal = TRUE;
  options[8].meaning = _T("Check that received data is correct");
  options[9].key = _T("i");
  options[9].type = Option_boolean;
  options[9].def.booleanVal = TRUE;
  options[9].meaning = _T("TRUE => Runs until user enters 'q'");
  options[10].key = _T("runtime");
  options[10].type = Option_uint;
  options[10].def.uintVal = 10;
  options[10].meaning = _T("If -i, run time in s (0 => for ever)");
  arguments.nOption = 11;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  parse_command_line(_T("DMA example application"), _T("dma"), argc, argv, &arguments);

  if (options[0].specified && options[1].specified) {
    _tprintf(_T("Do not specify both -card and -index\n"));
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Failed to open card with ID %lu: %s\n"), (unsigned long)cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  /* Allocate memory for the DMA data buffers */
  bufSize = options[2].value.uintVal;
  pRecvBuf = (uint8_t*)malloc(bufSize * 2);
  if (pRecvBuf == NULL) {
    _tprintf(_T("Failed to allocate buffers for threads\n"));
    ret = -1;
    goto done;
  }
  pSendBuf = pRecvBuf + bufSize;
  _tprintf(_T("sendbuf=%p recvbuf=%p\n"), pSendBuf, pRecvBuf);

  clockFreq = options[3].value.floatVal * 1.0e6f;
  bUse64bit = options[4].value.booleanVal;
  sendChan = options[5].value.uintVal;
  recvChan = options[6].value.uintVal;
  bDmaConfiguration = options[7].value.booleanVal;
  bCheckData = options[8].value.booleanVal;
  bInteractive = options[9].value.booleanVal;
  secondsToRun = options[10].value.uintVal;

#if defined(_WIN32)
  /*
  ** Create a manual reset event for use when configuring the FPGA
  ** using DMA.
  */
  hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
  if (hEvent == NULL) {
    _tprintf(_T("Failed to create Win32 Event\n"));
    ret = -1;
    goto done;
  }
#endif

  /* Get card information, for device fitted */
  status = ADMXRC2_GetCardInfo(hCard, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("Failed to get card info: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /* Ensure that DMA channels to use for sender and receiver threads are valid */
  if (sendChan == recvChan) {
    _tprintf(_T("*** Sender and receiver DMA channels cannot be the same\n"));
    ret = -1;
    goto done;
  }
  if (sendChan >= cardInfo.NumDMAChan) {
    _tprintf(_T("*** Invalid value (%lu) for sender DMA channel\n"), (unsigned long)sendChan);
    ret = -1;
    goto done;
  }
  if (recvChan >= cardInfo.NumDMAChan) {
    _tprintf(_T("*** Invalid value (%lu) for receiver DMA channel\n"), (unsigned long)recvChan);
    ret = -1;
    goto done;
  }

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
  case ADMXRC2_BOARD_ADMXRC5TZ:
  case ADMXRC2_BOARD_ADMXRC5TDA1:
    if (bUse64bit) {
      _tprintf(_T("*** The selected card cannot be run in 64-bit mode.\n"));
      ret = -1;
      goto done;
    }
    break;

  case ADMXRC2_BOARD_ADMPCIE6S1:
    if (bUse64bit) {
      _tprintf(_T("*** The selected card cannot be run in 64-bit mode.\n"));
      ret = -1;
      goto done;
    }
    break;

  case ADMXRC2_BOARD_ADMXPL:
  case ADMXRC2_BOARD_ADMXP:
  case ADMXRC2_BOARD_ADPXPI:
  case ADMXRC2_BOARD_ADMXRC4FX:
  case ADMXRC2_BOARD_ADPEXRC4FX:
  case ADMXRC2_BOARD_ADMXRC5LX:
  case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
  case ADMXRC2_BOARD_ADMXRC5T2ADV6:
  case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
    break;

  default:
    _tprintf(_T("This example must be run on one of the following cards:\n\n"));
    _tprintf(_T("\tADM-XRC\n"));
    _tprintf(_T("\tADM-XRC-P\n"));
    _tprintf(_T("\tADM-XRC-II-Lite\n"));
    _tprintf(_T("\tADM-XRC-II\n"));
    _tprintf(_T("\tADM-XPL\n"));
    _tprintf(_T("\tADM-XP\n"));
    _tprintf(_T("\tADP-WRC-II\n"));
    _tprintf(_T("\tADP-DRC-II\n"));
    _tprintf(_T("\tADP-XPI\n"));
    _tprintf(_T("\tADM-XRC-4LX\n"));
    _tprintf(_T("\tADM-XRC-4SX\n"));
    _tprintf(_T("\tADM-XRC-4FX / ADM-XMC-4FX\n"));
    _tprintf(_T("\tADPE-XRC-4FX\n"));
    _tprintf(_T("\tADM-XRC-5LX\n"));
    _tprintf(_T("\tADM-XRC-5T1\n"));
    _tprintf(_T("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6\n"));
    _tprintf(_T("\tADM-XRC-5TZ\n"));
    _tprintf(_T("\tADM-XRC-5T-DA1\n"));
    _tprintf(_T("\tADM-XRC-5T2-ADV-CC1\n"));
    _tprintf(_T("\tADM-PCIE-6S1\n"));
    ret = -1;
    goto done;
  }

  if (cardInfo.NumClock > 0) {
    status = ADMXRC2_SetClockRate(hCard, ADMXRC2_CLOCK_LCLK, clockFreq, NULL);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Failed to set LCLK to %.1fMHz: %s\n"), clockFreq / 1.0e6f, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  status = ADMXRC2_GetSpaceInfo(hCard, 0, &spaceInfo);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("Failed to get information about FPGA space\n"));
    ret = -1;
    goto done;
  }
  pFpgaSpace = (volatile uint32_t*)spaceInfo.VirtualBase;
  if (pFpgaSpace == NULL) {
    _tprintf(_T("*** Failed to retrieve pointer to FPGA space\n"));
    return -1;
  }

  if (bUse64bit) {
    uint32_t config;

    config = ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_WIDTH_64;
    status = ADMXRC2_SetSpaceConfig(hCard, 0, config);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("*** Failed to set space configuration: %s\n"), ADMXRC2_GetStatusString(status));
      return -1;
    }
  }

  /* Load the FPGA using DMA and measure the configuration time */
  if (1) {
    const int num_rep = 5;
    int i;
    ADMXRC2_IMAGE image;
    unsigned long size;
    admxrc2_time_t t;

    if (bUse64bit) {
      filename = get_fpga_filename(&cardInfo, _T("ddma64"), &arguments);
    } else {
      filename = get_fpga_filename(&cardInfo, _T("ddma"), &arguments);
    }
    status = ADMXRC2_LoadBitstream(hCard, filename, &image, &size);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf (_T("Failed to load FPGA file '%s': %s\n"), filename, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }

    t = timeGet();
    for (i = 0; i < num_rep; i++) {
      if (bDmaConfiguration) {
        status = ADMXRC2_ConfigureFromBufferDMA(hCard, image, size, ADMXRC2_DMACHAN_ANY, hEvent);
        if (status != ADMXRC2_SUCCESS) {
          ADMXRC2_UnloadBitstream(image);
          _tprintf(_T("Failed to configure FPGA from buffer with DMA: %s\n"), ADMXRC2_GetStatusString(status));
          ret = -1;
          goto done;
        }
      } else {
        status = ADMXRC2_ConfigureFromBuffer(hCard, image, size);
        if (status != ADMXRC2_SUCCESS) {
          ADMXRC2_UnloadBitstream(image);
          _tprintf(_T("Failed to configure FPGA from buffer: %s\n"), ADMXRC2_GetStatusString(status));
          ret = -1;
          goto done;
        }
      }
    }
    t = timeSubtract(timeGet(), t);

    _tprintf(_T("FPGA configuration time = %.1fms\n"), timeToDouble(t) / (double)i);

    ADMXRC2_UnloadBitstream(image);
  }

  /*
  ** Build the DMA mode word. We want bursting, demand-mode,
  ** use ready, use bterm and fixed local bus addressing.
  */
  mode = ADMXRC2_BuildDMAModeWord(
    cardInfo.BoardType,
    bUse64bit ? ADMXRC2_IOWIDTH_64 : ADMXRC2_IOWIDTH_32,
    0,
    ADMXRC2_DMAMODE_FIXEDLOCAL |
    ADMXRC2_DMAMODE_USEREADY |
    ADMXRC2_DMAMODE_USEBTERM |
    ADMXRC2_DMAMODE_BURSTENABLE |
    ADMXRC2_DMAMODE_DEMAND);

#if defined(_WIN32)
  /*
  ** A semaphore is required for the send thread to notify the receive thread
  ** that the receive thread can proceed with its DMA.
  */
  readySem = CreateSemaphore(NULL, 0, 1024, NULL);
  if (readySem == NULL) {
    _tprintf(_T("Failed to create Win32 Semaphore\n"));
    ret = -1;
    goto done;
  }
#else
  sem_init(&readySem, FALSE, 0);
  bReadySemInited = TRUE;
#endif

  /* Get a DMA descriptor for the buffer used for DMA transfers to the card */
  status = ADMXRC2_SetupDMA(hCard, pSendBuf, bufSize, 0, &hSendBuf);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("Failed to set up send DMA buffer: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  bSendBufValid = TRUE;

  /* Get a DMA descriptor for the buffer used for DMA transfers from the card */
  status = ADMXRC2_SetupDMA(hCard, pRecvBuf, bufSize, 0, &hRecvBuf);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("Failed to set up receive DMA buffer: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  bRecvBufValid = TRUE;

  /*
  ** Start the DMA threads.
  */
  bTerminate = FALSE;
  bTerminateRecv = FALSE;
  sent = 0.0;
  received = 0.0;
  _tprintf(_T("Starting thread(s)...\n"));
  start = timeGet();
#if defined(_WIN32)
  hSendThread = CreateThread(NULL, 16384, sendThreadFunc, NULL, 0, &sendThreadId);
  hRecvThread = CreateThread(NULL, 16384, recvThreadFunc, NULL, 0, &recvThreadId);
#else
  pthread_create(&pSendThread, NULL, sendThreadFunc, NULL);
  pthread_create(&pRecvThread, NULL, recvThreadFunc, NULL);
#endif

  sleepMs(200);

  /* Display average transfer rates for ever or until user enters 'q' */
  while (1) {
    int c;
    double totalM, sentM, receivedM;
    admxrc2_time_t now;
    double elapsed;

    now = timeGet();
    elapsed = timeToDouble(timeSubtract(now, start));
    sentM = sent / 1048576.0;
    receivedM = received / 1048576.0;
    totalM = sentM + receivedM;

    _tprintf(_T("Sent %.2f MiB, received %.2f MiB, total %.2f MiB\n"), sentM, receivedM, totalM);
    if (elapsed > 0.001) {
      _tprintf(_T("Send rate %.2f MiB/s, receive rate %.2f MiB/s, total rate %.2f MiB/s\n"),
        receivedM / elapsed, receivedM / elapsed, totalM / elapsed);
    }

    if (bInteractive) {
      _tprintf(_T("Enter 'q' to exit, anything else to view rates:\n"));
      c = _gettchar();
      if (c == _T('q')) {
        break;
      }
      sleepMs(100);
    } else {
      sleepMs(1000);
      if (0 != secondsToRun) {
        unsigned int secondsElapsed = (unsigned int)timeToDouble(timeSubtract(timeGet(), start));
        if (secondsElapsed >= secondsToRun) {
          break;
        }
      }
    }
  }

  /*
  ** Terminate the threads.
  ** First, terminate the sender DMA thread.
  ** Second, terminate the receiver DMA thread. We increment the semaphore once
  ** in case the receiver DMA thread is currently waiting on the semaphore.
  */
  _tprintf(_T("Terminating threads...\n"));
  bTerminate = TRUE;
#if defined(_WIN32)
  if (WaitForSingleObject(hSendThread, INFINITE) != WAIT_OBJECT_0) {
    if (!TerminateThread(hSendThread, 0)) {
      _tprintf(_T("Failed to terminate send thread\n"));
    }
  }
#else
  pthread_join(pSendThread, NULL);
#endif
  bTerminateRecv = TRUE;
#if defined(_WIN32)
  ReleaseSemaphore(readySem, 1, NULL);
  if (WaitForSingleObject(hRecvThread, INFINITE) != WAIT_OBJECT_0) {
    if (!TerminateThread(hRecvThread, 0)) {
      _tprintf(_T("Failed to terminate sucker thread\n"));
    }
  }
#else
  sem_post(&readySem);
  pthread_join(pRecvThread, NULL);
#endif

done:
  /* Free the DMA descriptors */
  if (bSendBufValid) {
    status = ADMXRC2_UnsetupDMA(hCard, hSendBuf);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Failed to delete DMA send buffer: %s\n"), ADMXRC2_GetStatusString(status));
    }
  }
  if (bRecvBufValid) {
    status = ADMXRC2_UnsetupDMA(hCard, hRecvBuf);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Failed to delete DMA receive buffer: %s\n"), ADMXRC2_GetStatusString(status));
    }
  }

#if defined(_WIN32)
  /* Free the Win32 Semaphore */
  if (readySem != NULL) {
    CloseHandle(readySem);
  }

  /* Free the Win32 Event */
  if (hEvent != NULL) {
    CloseHandle(hEvent);
  }
#else
  if (bReadySemInited) {
    sem_destroy(&readySem);
  }
#endif

  /* Close the device */
  if (hCard != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(hCard);
  }
  
  /* Free the memory used for the DMA data buffers */
  if (pRecvBuf != NULL) {
    free(pRecvBuf);
  }

  return ret;
}
