/*
** itest.c - example demonstrating interrupt wait functionality in
**           the ADM-XRC API via Win32 events.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**   o ADM-XRC-5T2-ADV-CC1
**   o ADM-PCIE-6S1
**
** (C) Copyright 2002-2014 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <tchar.h>

#include <admxrc2.h>

#include <sdk_common.h>

/* FPGA registers */
#define FPGA_REG_IMASK (0x0U)
#define FPGA_REG_ISTAT (0x4U)
#define FPGA_REG_IARM	 (0x8U)
#define FPGA_REG_TEST	 (0xcU)

static ADMXRC2_HANDLE	hCard = ADMXRC2_HANDLE_INVALID_VALUE;

static unsigned long numInterruptSeen;
static HANDLE hEvent = NULL;
static BOOLEAN bEventReged = FALSE;

static HANDLE hInterruptThread;
static uint32_t interruptThreadId;
static bool_t bInterruptThreadRunning = FALSE;
static bool_t bTerminate = FALSE;

static DWORD WINAPI
interruptThread(
  void* pArg)
{
  DWORD result;
  uint32_t istat, dummy;

  _tprintf(_T("Interrupt thread started\n"));
  
  while (1) {
    result = WaitForSingleObject(hEvent, INFINITE);
    if (result != WAIT_OBJECT_0) {
      DWORD lastError = GetLastError();
      _tprintf(_T("*** Interrupt thread failed to wait for event (Win32 error 0x%08lX)\n"), lastError, lastError);
      break;
    }

    if (bTerminate) {
      break;
    }

    numInterruptSeen++;

    /* Get the pending FPGA interrupts */
    ADMXRC2_Read(hCard, ADMXRC2_IOWIDTH_32, ADMXRC2_IOLOCAL, FPGA_REG_ISTAT, &istat, sizeof(istat));

    /* Clear the pending FPGA interrupts */
    ADMXRC2_Write(hCard, ADMXRC2_IOWIDTH_32, ADMXRC2_IOLOCAL, FPGA_REG_ISTAT, &istat, sizeof(istat));

    /* Rearm the FINTI_L line by forcing it high for 1 cycle */
    ADMXRC2_Write(hCard, ADMXRC2_IOWIDTH_32, ADMXRC2_IOLOCAL, FPGA_REG_IARM, &dummy, sizeof(dummy));
  }

  return 0;
}

int main(int argc, TCHAR *argv[])
{
  const double clockFreq = 33.0e6;
  int ret = 0;
  ADMXRC2_STATUS status;
  ADMXRC2_CARD_INFO cardInfo;
  unsigned long numInterruptGenerated;
  const TCHAR* filename;
  Arguments arguments;
  Option options[2];
  uint32_t val32;

  options[0].key = _T("card");
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("ID of card to use");
  options[1].key = _T("index");
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("index of card to use");
  arguments.nOption = 2;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  parse_command_line(_T("ITEST example application"), _T("itest"), argc, argv, &arguments);

  if (options[0].specified && options[1].specified) {
    _tprintf(_T("Do not specify both -card and -index\n"));
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("*** Failed to open card with ID %lu: %s\n"), (unsigned long)cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  status = ADMXRC2_GetCardInfo(hCard, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  
  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADMXPL:
  case ADMXRC2_BOARD_ADMXP:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
  case ADMXRC2_BOARD_ADPXPI:
  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
  case ADMXRC2_BOARD_ADMXRC4FX:
  case ADMXRC2_BOARD_ADPEXRC4FX:
  case ADMXRC2_BOARD_ADMXRC5LX:
  case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMXRC5TZ:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
  case ADMXRC2_BOARD_ADMXRC5T2ADV6:
  case ADMXRC2_BOARD_ADMXRC5TDA1:
  case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
  case ADMXRC2_BOARD_ADMPCIE6S1:
    break;

  default:
    _tprintf(_T("This example must be run on one of the following cards:\n\n"));
    _tprintf(_T("\tADM-XRC\n"));
    _tprintf(_T("\tADM-XRC-P\n"));
    _tprintf(_T("\tADM-XRC-II-Lite\n"));
    _tprintf(_T("\tADM-XRC-II\n"));
    _tprintf(_T("\tADM-XPL\n"));
    _tprintf(_T("\tADM-XP\n"));
    _tprintf(_T("\tADP-WRC-II\n"));
    _tprintf(_T("\tADP-DRC-II\n"));
    _tprintf(_T("\tADP-XPI\n"));
    _tprintf(_T("\tADM-XRC-4LX\n"));
    _tprintf(_T("\tADM-XRC-4SX\n"));
    _tprintf(_T("\tADM-XRC-4FX / ADM-XMC-4FX\n"));
    _tprintf(_T("\tADPE-XRC-4FX\n"));
    _tprintf(_T("\tADM-XRC-5LX\n"));
    _tprintf(_T("\tADM-XRC-5T1\n"));
    _tprintf(_T("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6\n"));
    _tprintf(_T("\tADM-XRC-5TZ\n"));
    _tprintf(_T("\tADM-XRC-5T-DA1\n"));
    _tprintf(_T("\tADM-XRC-5T2-ADV-CC1\n"));
    _tprintf(_T("\tADM-PCIE-6S1\n"));
    ret = -1;
    goto done;
  }

  if (cardInfo.NumClock > 0) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC2_SetClockRate(hCard, ADMXRC2_CLOCK_LCLK, clockFreq, NULL);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("*** Failed to set LCLK to %.1fMHz: %s\n"), clockFreq / 1.0e6f, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  filename = get_fpga_filename(&cardInfo, _T("itest"), &arguments);
  status = ADMXRC2_ConfigureFromFile(hCard, filename);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to load the bitstream '%s': %s\n"), filename, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /* Create an auto-reset event */
  hEvent = CreateEvent(NULL, FALSE, FALSE,	NULL);
  if (hEvent == NULL) {
    _tprintf(_T("*** Failed to create Win32 event\n"));
    ret = -1;
    goto done;
  }

  /*
  ** Register the event with the card, so that when an FPGA interrupt occurs,
  ** the event will be signalled.
  */
  status = ADMXRC2_RegisterInterruptEvent(hCard, hEvent);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to register interrupt event: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  bEventReged = TRUE;

  /* Start up the thread that waits for interrupts. */
  numInterruptSeen = 0;
  bTerminate = FALSE;
  hInterruptThread = CreateThread(NULL, 0, interruptThread, NULL, 0, &interruptThreadId);
  if (hInterruptThread == NULL) {
    _tprintf(_T("*** Failed to start interrupt thread\n"));
    ret = -1;
    goto done;
  }
  bInterruptThreadRunning = TRUE;

  /* Enable interrupts in the FPGA */
  val32 = 0x0U;
  status = ADMXRC2_Write(hCard, ADMXRC2_IOWIDTH_32, ADMXRC2_IOLOCAL, FPGA_REG_IMASK, &val32, sizeof(val32));
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("Write to FPGA_IMASK failed: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /*
  ** Allow the user to generate interrupts by pressing enter.
  ** Ctrl-Z exits this loop.
  */
  numInterruptGenerated = 0;
  while (1) {
    int c;

    _tprintf(_T("Enter 'q' to quit, or anything else to generate an interrupt:\n"));
    c = getchar();
    if (c == _TEOF || _totlower(c) == _T('q')) {
      break;
    }

    val32 = 0x1U;
    status = ADMXRC2_Write(hCard, ADMXRC2_IOWIDTH_32, ADMXRC2_IOLOCAL, FPGA_REG_TEST, &val32, sizeof(val32));
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Write to FPGA_TEST failed: %s\n"), ADMXRC2_GetStatusString(status));
    }
    numInterruptGenerated++;
  }

  _tprintf(_T("Generated %lu interrupts\n"), (unsigned long)numInterruptGenerated);
  _tprintf(_T("Interrupt thread saw %lu interrupt(s)\n"), (unsigned long)numInterruptSeen);

done:
  if (bInterruptThreadRunning) {
    /* Force the interrupt thread to terminate */
    bTerminate = TRUE;
    SetEvent(hEvent);
    /* Wait for the interrupt thread to terminate */
    WaitForSingleObject(hInterruptThread, 1000);
    CloseHandle(hInterruptThread);
  }

  if (hEvent != NULL) {
    if (bEventReged) {
      /*
      ** Unregister the event, so that it will no longer be signalled
      ** on an FPGA interrupt.
      */
      status = ADMXRC2_UnregisterInterruptEvent(hCard, hEvent);
      if (status != ADMXRC2_SUCCESS) {
        _tprintf(_T("*** Failed to unregister interrupt event: %s\n"), ADMXRC2_GetStatusString(status));
      }
    }
    CloseHandle(hEvent);
  }

  if (hCard != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(hCard);
  }
  
  return ret;
}
