/*
** memoryf.c - Example demonstrating how to perform chip-driven memory test
**             on an Alpha Data reconfigurable computing card.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XP
**   o ADM-XPL
**   o ADP-DRC-II
**   o ADP-WRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**   o ADM-XRC-5T2-ADV-CC1
**
** (C) Copyright 2008-2014 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _tmain main
# define _T(c) c
# define _tprintf printf

#endif

#include <admxrc2.h>

#include <sdk_common.h>
#include <platform_endian.h>
#include <platform_time.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define minimumOf2Macro(a, b) ((a) < (b) ? (a) : (b))

/* If nonzero, host reads memory during memory test to try to provoke some kind of error */
#define HOST_READ_ACTIVITY (1)

/*
** Register indices and bit definitions
*/

#define BANK_REG                        (0x0)               /* Bank select register */
#define BANK_REG_BANKMASK               (0xfU << 0)

#define PAGE_REG                        (0x1)               /* Page register */
#define PAGE_REG_PAGEMASK               (0x7ffU << 0)

#define MEMCTL_REG                      (0x2)               /* Memory control register */
#define MEMCTL_REG_MEMRST               (0x1U << 0)

#define STATUS_REG                      (0x4)               /* Status register */
#define STATUS_REG_LCLKLOCKED           (0x1U << 0)
#define STATUS_REG_LCLKSTICKY           (0x1U << 1)
#define STATUS_REG_SHIFT_LOCKED         (8)
#define STATUS_REG_SHIFT_STICKY         (16)

#define MEMSTAT_REG                     (0x6)               /* Memory status register */
#define MEMSTAT_REG_SHIFT_TRAINED       (0)

#define MODEx_REG(n)                    (0x10 + (n))        /* Bank 'n' mode register */
#define MODEx_REG_DDR2SSRAM_BLMASK      (0x3U << 0)         /* DDR2 SSRAM burst length mask */
#define MODEx_REG_DDR2SSRAM_BL2         (0x0U << 0)         /* DDR2 SSRAM burst length 2 */
#define MODEx_REG_DDR2SSRAM_BL4         (0x1U << 0)         /* DDR2 SSRAM burst length 4 */
#define MODEx_REG_DDR2SSRAM_DLLON       (0x0U << 2)         /* DDR2 SSRAM DLL enable */
#define MODEx_REG_DDR2SSRAM_DLLOFF      (0x1U << 2)         /* DDR2 SSRAM DLL disable */

#define MODEx_REG_ZBTSSRAM_FLOWTHRU     (0x0U << 0)         /* ZBT SSRAM pipelined mode disable */
#define MODEx_REG_ZBTSSRAM_PIPELINE     (0x1U << 0)         /* ZBT SSRAM pipelined mode enable */

#define MODEx_REG_DDRSDRAM_UNBUFFERED   (0x0U << 0)
#define MODEx_REG_DDRSDRAM_REGISTERED   (0x1U << 0)
#define MODEx_REG_DDRSDRAM_X8X16        (0x0U << 1)
#define MODEx_REG_DDRSDRAM_X4           (0x1U << 1)
#define MODEx_REG_DDRSDRAM_ROWBITS_12   (0x0U << 2)
#define MODEx_REG_DDRSDRAM_ROWBITS_13   (0x1U << 2)
#define MODEx_REG_DDRSDRAM_ROWBITS_14   (0x2U << 2)
#define MODEx_REG_DDRSDRAM_ROWBITS_15   (0x3U << 2)
#define MODEx_REG_DDRSDRAM_COLBITS_M4   (0x0U << 4)         /* col bits = row bits - 4 */
#define MODEx_REG_DDRSDRAM_COLBITS_M3   (0x1U << 4)         /* col bits = row bits - 3 */
#define MODEx_REG_DDRSDRAM_COLBITS_M2   (0x2U << 4)         /* col bits = row bits - 2 */
#define MODEx_REG_DDRSDRAM_COLBITS_M1   (0x3U << 4)         /* col bits = row bits - 1 */
#define MODEx_REG_DDRSDRAM_BANKBITS_0   (0x0U << 6)
#define MODEx_REG_DDRSDRAM_BANKBITS_1   (0x1U << 6)
#define MODEx_REG_DDRSDRAM_BANKBITS_2   (0x2U << 6)
#define MODEx_REG_DDRSDRAM_BANKBITS_3   (0x3U << 6)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_1  (0x0U << 8)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_2  (0x1U << 8)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_4  (0x2U << 8)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_8  (0x3U << 8)

#define MODEx_REG_DDR2SDRAM_TRFC_1Gb    (0x0U << 10)        /* Refresh cycle time for 1Gb DDR-II SDRAM part */
#define MODEx_REG_DDR2SDRAM_TRFC_256Mb  (0x1U << 10)        /* Refresh cycle time for 256Mb DDR-II SDRAM part */
#define MODEx_REG_DDR2SDRAM_TRFC_512Mb  (0x2U << 10)        /* Refresh cycle time for 512Mb DDR-II SDRAM part */
#define MODEx_REG_DDR2SDRAM_TRFC_2Gb    (0x3U << 10)        /* Refresh cycle time for 2Gb DDR-II SDRAM part */

#define SPD_REG                         (0x5)
#define SPD_REG_DO                      (0x1U << 31)
#define SPD_REG_BUSY                    (0x1U << 31)
#define SPD_REG_SHIFT_ADDR              (0)
#define SPD_REG_SHIFT_WRDATA            (8)
#define SPD_REG_SHIFT_RDDATA            (16)
#define SPD_REG_ERR                     (0x1U << 24)
#define SPD_REG_SHIFT_SLOT              (25)

#define USER_REG(x)                     (0x40U | (x))

/* Max. number of memory banks supported by this application */
#define MAX_BANK (16)

/* Max. logical width of a memory bank, in bytes */
#define MAX_BANK_WIDTH (64)

struct _Globals {
  void* ppFpgaSpace;
  volatile uint32_t* pFpgaRegs;
  volatile uint32_t* pMemory;
  ADMXRC2_HANDLE hCard;
  uint32_t bankTestMask;
  bool_t bBankDdr[MAX_BANK];
  unsigned int bankWidth[MAX_BANK];
  uint64_t bankSize[MAX_BANK];
  bool_t bUse64Bit;
};

typedef struct _Globals Globals;

static bool_t
readSpdByte(
  Globals* pGlob,
  unsigned int slot,
  unsigned int index,
  uint8_t* value)
{
  volatile uint32_t* regs = pGlob->pFpgaRegs;
  uint32_t val32;
  uint32_t stat32;
  bool_t bFail;

  val32 = (index & 0xffU) << SPD_REG_SHIFT_ADDR;
  val32 |= (slot & 0x7U) << SPD_REG_SHIFT_SLOT;
  regs[SPD_REG] = val32;
  val32 |= SPD_REG_DO;
  regs[SPD_REG] = val32;
  stat32 = regs[SPD_REG];
  while (1) {
    stat32 = regs[SPD_REG];
    if (!(stat32 & SPD_REG_BUSY)) {
      break;
    }
  }
  if (stat32 & SPD_REG_ERR) {
    bFail = TRUE;
  } else {
    bFail = FALSE;
  }
  *value = (uint8_t) (stat32 >> SPD_REG_SHIFT_RDDATA);
  return bFail;
}

static bool_t
readSpd(
  Globals* pGlob,
  unsigned int slot,
  uint8_t data[128])
{
  unsigned int i;
  bool_t bFail;

  for (i = 0; i < 128; i++) {
    bFail = readSpdByte(pGlob, slot, i, &data[i]);
    if (bFail || (i == 0 && data[i] == 0xffU)) {
      break;
    }
  }
  return (!bFail && data[0] != 0xffU);
}

static void
selectBank(
  Globals* pGlob,
  unsigned int bank)
{
  pGlob->pFpgaRegs[BANK_REG] = (uint32_t)bank & BANK_REG_BANKMASK;
  pGlob->pFpgaRegs[BANK_REG];
}

/*
** Check that local bus clock DCM has achieved lock.
*/
static bool_t
waitForLclkDcm(
  Globals* pGlob)
{
  uint32_t status;
  uint32_t val32;

  /*
  ** Delay for 500ms to allow DLLs/DCMs time to lock, and then verify
  ** that they are locked as expected.
  */
  sleepMs(500);
  status = pGlob->pFpgaRegs[STATUS_REG];
  _tprintf(_T("STATUS=0x%08lX\n"), (unsigned long)status);
  if (!(status & STATUS_REG_LCLKLOCKED)) {
    _tprintf(_T("*** LCLK DCM not locked\n"));
    return FALSE;
  }
  /* Clear "sticky" loss of lock bit */
  val32 = STATUS_REG_LCLKSTICKY;
  pGlob->pFpgaRegs[STATUS_REG] = val32;
  pGlob->pFpgaRegs[STATUS_REG];

  return TRUE;
}

/*
** Check that DCMs/PLLs in memory clock domain have achieved lock, and that
** any memory banks that are self-training have trained themselves without
** error.
*/
static bool_t
checkForMemoryLock(
  Globals* pGlob,
  unsigned int numLockFlags)
{
  uint32_t status, memstat, mask, val32;

  mask = (1 << numLockFlags) - 1;
  status = pGlob->pFpgaRegs[STATUS_REG];
  memstat = pGlob->pFpgaRegs[MEMSTAT_REG];
  _tprintf(_T("STATUS=0x%08lX MEMSTAT=0x%08lX\n"),
    (unsigned long)status, (unsigned long)memstat);
  val32 = status >> STATUS_REG_SHIFT_LOCKED;
  val32 &= mask;
  if (val32 != mask) {
    _tprintf(_T("*** Not all lock flags asserted: expected=0x%lX actual=0x%lX\n"),
      (unsigned long)mask, (unsigned long)val32);
    return FALSE;
  }
  val32 = memstat >> MEMSTAT_REG_SHIFT_TRAINED;
  val32 &= pGlob->bankTestMask;
  if (val32 != pGlob->bankTestMask) {
    _tprintf(_T("*** Not all memory banks trained: expected=0x%lX actual=0x%lX\n"),
      (unsigned long)pGlob->bankTestMask, (unsigned long)val32);
    return FALSE;
  }
  /* Clear sticky loss-of-lock / loss-of-trained bits */
  val32 = 0xffU << STATUS_REG_SHIFT_STICKY;
  pGlob->pFpgaRegs[STATUS_REG] = val32;
  pGlob->pFpgaRegs[STATUS_REG];

  return TRUE;
}

/*
** Check that DLLs/DCMs/PLLs in memory clock domain did not lose lock during
** the test, and that memory banks did not transition to untrained.
*/
static bool_t
checkStickyStatus(
    Globals* pGlob,
    unsigned int numLockFlags)
{
    uint32_t status, memstat, mask, val32;

    mask = (1 << numLockFlags) - 1;
    status = pGlob->pFpgaRegs[STATUS_REG];
    memstat = pGlob->pFpgaRegs[MEMSTAT_REG];
    _tprintf(_T("STATUS=0x%08lX MEMSTAT=0x%08lX\n"),
        (unsigned long)status, (unsigned long)memstat);
    val32 = status >> STATUS_REG_SHIFT_STICKY;
    val32 &= mask;
    if (val32) {
        _tprintf(_T("*** Loss of lock occurred: flags=0x%lX\n"), (unsigned long)val32);
        return FALSE;
    }

    return TRUE;
}

static bool_t
getDimmMode(
  ADMXRC2_CARD_INFO* pCardInfo,
  ADMXRC2_BANK_INFO* pBankInfo,
  Globals* pGlob,
  unsigned int bank,
  uint32_t* pMode,
  bool_t* x4)
{
  uint8_t spd[128];
  bool_t bPresent;
  unsigned long size;
  unsigned int rows, cols, banks, pbanks;
  unsigned int devWidth;
  uint8_t clMask;
  uint8_t modAttr;
  bool_t bRegistered;
  uint32_t mode = 0;

  bPresent = readSpd(pGlob, bank, spd);
  if (bPresent) {
    rows = spd[3];      /* Number of bits used for row address on devices */
    cols = spd[4];      /* Number of bits used for column address on devices */
    pbanks = spd[5];    /* Number of physical banks on DIMM */
    devWidth = spd[13]; /* DQ width of devices */
    banks = spd[17];    /* Number of banks on devices */
    clMask = spd[18];   /* CAS latencies supported by devices */
    modAttr = spd[21];  /* DIMM module attributes */
    switch (devWidth) {
    case 4:
    case 8:
    case 16:
      break;
    default:
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): unknown device data width (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)devWidth);
      return FALSE;
    }
    if (banks < 1 || banks > 8) {
      _tprintf(_T(" *** Bank %lu (DDR SDRAM): unsupported device banks value (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)banks);
      return FALSE;
    }
    if (!(clMask & 0x8U)) {
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): CAS latency 2.5 not supported by devices, not testing\n"),
        (unsigned long)bank);
      return FALSE;
    }
    switch (modAttr) {
    case 0x20: /* Unbuffered */
      bRegistered = FALSE;
      break;
    case 0x26: /* Registered */
      bRegistered = TRUE;
      break;
    default:
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid DIMM module attributes value (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)modAttr);
      return FALSE;
    }
    switch (rows) {
    case 12:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_12;
      break;
    case 13:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
      break;
    case 14:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_14;
      break;
    case 15:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_15;
      break;
    default:
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid number of row bits (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)rows);
      return FALSE;
    }
    switch (cols + 4 - rows) {
    case 0:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M4;
      break;
    case 1:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
      break;
    case 2:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M2;
      break;
    case 3:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M1;
      break;
    default:
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid number of column bits (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)cols);
      return FALSE;
    }
    switch (banks) {
    case 1:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_0;
      break;
    case 2:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_1;
      break;
    case 4:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
      break;
    case 8:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_3;
      break;
    default:
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid number of banks (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)banks);
      return FALSE;
    }
    switch (pbanks) {
    case 1:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
      break;
    case 2:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_2;
      break;
    case 4:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_4;
      break;
    case 8:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_8;
      break;
    default:
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid number of physical banks (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)pbanks);
      return FALSE;
    }
    if (bRegistered) {
      mode |= 0x1U;
    }
    if (devWidth == 4) {
      mode |= 0x2U;
    }
    pBankInfo[bank].Size = (1 << (rows + cols - 1)) * banks * pbanks;
    size = (pBankInfo[bank].Size >> 20) * (pBankInfo[bank].Width >> 3);
    _tprintf(_T(" Bank %lu (DDR SDRAM): x%lu %s rows=%lu cols=%lu banks=%lu physBanks=%lu %lu MiB\n"),
      (unsigned long)bank,
      (unsigned long)devWidth,
      bRegistered ? "registered" : "unbuffered",
      (unsigned long)rows, (unsigned long)cols,
      (unsigned long)banks, (unsigned long)pbanks, (unsigned long)size);
  } else {
    _tprintf(_T(" +++ Bank %lu (DDR SDRAM): SPD data not found, not testing\n"),
      (unsigned long)bank);
    return FALSE;
  }

  /* Bank SPD seems OK */
  *pMode = mode;
  *x4 = (devWidth == 4);
  return TRUE;
}

static void
setMemoryPortConfiguration(
  Globals* pGlob,
  uint32_t* modes)
{
  unsigned int i;

  _tprintf(_T("Configuring memory ports...\n"));
  for (i = 0; i < MAX_BANK; i++) {
    if (!(pGlob->bankTestMask & (0x1UL << i))) {
      continue;
    }
    pGlob->pFpgaRegs[MODEx_REG(i)] = modes[i];
  }
}

static void
scanMemoryPorts(
  ADMXRC2_CARD_INFO* pCardInfo,
  ADMXRC2_BANK_INFO* pBankInfo,
  Globals* pGlob,
  uint32_t* modes)
{
  unsigned int i;
  bool_t dimmX4[4];

  /* Initialize memory bank config registers where necessary */
  _tprintf(_T("Scanning memory ports...\n"));
  for (i = 0; i < pCardInfo->NumRAMBank; i++) {
    if (!(pGlob->bankTestMask & (0x1U << i))) {
      continue;
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_SRAM_DDR2) {
      /* Select BL4 for DDR-II SSRAM and ensure DLL is enabled */
      modes[i] = MODEx_REG_DDR2SSRAM_BL4 | MODEx_REG_DDR2SSRAM_DLLON;
      pGlob->bBankDdr[i] = TRUE;
      _tprintf(_T(" Bank %lu (DDR-II SRAM): burst length 4, DLL enabled\n"),
        (unsigned long)i);
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_ZBTP) {
      /* Select pipelined mode for ZBT SSRAM */
      modes[i] = MODEx_REG_ZBTSSRAM_PIPELINE;
      pGlob->bBankDdr[i] = FALSE;
      _tprintf(_T(" Bank %lu (ZBT SSRAM): pipelined mode\n"),
        (unsigned long)i);
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_ZBTFT) {
      /* Select flowthrough mode for ZBT SSRAM */
      modes[i] = MODEx_REG_ZBTSSRAM_FLOWTHRU;
      pGlob->bBankDdr[i] = FALSE;
      _tprintf(_T(" Bank %lu (ZBT SSRAM): flowthrough mode\n"),
        (unsigned long)i);
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_SDRAM_DDR) {
      uint32_t mode;

      /* Select registered, row bits, col bits, bank bits and phys bank bits for DDR SDRAM */
      switch (pCardInfo->BoardType) {
      case ADMXRC2_BOARD_ADMXPL:
      case ADMXRC2_BOARD_ADMXP:
        mode = 0;
        if (0x800000 == pBankInfo[i].Size) {
          /* unbuffered, row bits 13, col bits 9, bank bits 2, phys banks 1 */
          mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
          mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
          mode |= MODEx_REG_DDRSDRAM_COLBITS_M4;
          mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
          mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
          _tprintf(_T(" Bank %lu (DDR SDRAM): unbuffered, row 13, col 9, bank 2, phys banks 1\n"),
            (unsigned long)i);
        } else if (0x1000000 == pBankInfo[i].Size) {
          /* unregistered, row bits 13, col bits 10, bank bits 2, phys banks 1 */
          mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
          mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
          mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
          mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
          mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
          _tprintf(_T(" Bank %lu (DDR SDRAM): unbuffered, row 13, col 10, bank 2, phys banks 1\n"),
            (unsigned long)i);
        } else if (0x2000000 == pBankInfo[i].Size) {
          /* unregistered, row bits 13, col bits 11, bank bits 2, phys banks 1 */
          mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
          mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
          mode |= MODEx_REG_DDRSDRAM_COLBITS_M2;
          mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
          mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
          _tprintf(_T(" Bank %lu (DDR SDRAM): unbuffered, row 13, col  11, bank 2, phys banks 1\n"),
            (unsigned long)i);
        } else {
          pGlob->bankTestMask &= ~(0x1U << i);
          _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid size (0x%lX words), not testing\n"),
            (unsigned long)i, (unsigned long)pBankInfo[i].Size);
        }
        modes[i] = mode;
        pGlob->bBankDdr[i] = TRUE;
        break;

      case ADMXRC2_BOARD_ADPXPI:
        if (!getDimmMode(pCardInfo, pBankInfo, pGlob, i, &mode, &dimmX4[i])) {
          pGlob->bankTestMask &= ~(0x1U << i);
        }
        modes[i] = mode;
        pGlob->bBankDdr[i] = TRUE;
        break;
#if ZAP
      case ADMXRC2_BOARD_ADMXRC4SX:
      case ADMXRC2_BOARD_ADMXRC4LX:
        mode = 0;
        if (0x1000000 == pBankInfo[i].Size) {
          /* unregistered, row bits 13, col bits 9, bank bits 2, phys banks 1 */
          mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
          mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
          mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
          mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
          mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
          _tprintf(_T(" Bank %lu (DDR SDRAM): unbuffered, row 13, col 10, bank 2, phys banks 1\n"),
            (unsigned long)i);
        } else {
          pGlob->bankTestMask &= ~(0x1U << i);
          _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid size (0x%lX words), not testing\n"),
            (unsigned long)i, (unsigned long)pBankInfo[i].Size);
        }
        modes[i] = mode;
        pGlob->bBankDdr[i] = TRUE;
        break;
#endif /* ZAP */

      default:
        assert(FALSE); /* Shouldn't get here */
        break;
      }
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_SDRAM_DDR2) {
      uint32_t mode;

      /* Select registered, row bits, col bits, bank bits and phys bank bits for DDR SDRAM */
      switch (pCardInfo->BoardType) {
      case ADMXRC2_BOARD_ADMXRC4FX:
      case ADMXRC2_BOARD_ADPEXRC4FX:
      case ADMXRC2_BOARD_ADMXRC5LX:
      case ADMXRC2_BOARD_ADMXRC5T1:
      case ADMXRC2_BOARD_ADMXRC5T2:
      case ADMXRC2_BOARD_ADMXRC5T2ADV:
      case ADMXRC2_BOARD_ADMXRC5T2ADV6:
      case ADMXRC2_BOARD_ADMXRC5TDA1:
      case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
        mode = 0;
        if (0x1000000 == pBankInfo[i].Size) {
          /* unbuffered, row bits 13, col bits 9, bank bits 2, phys banks 1, Trfc for 1Gb part */
          mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
          mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
          mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
          mode |= MODEx_REG_DDRSDRAM_BANKBITS_3;
          mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
          mode |= MODEx_REG_DDR2SDRAM_TRFC_1Gb;
          _tprintf(_T(" Bank %lu (DDR-II SDRAM): unbuffered, row 13, col 10, bank 3, phys banks 1\n"),
            (unsigned long)i);
        } else if (0x2000000 == pBankInfo[i].Size) {
          /* unbuffered, row bits 14, col bits 10, bank bits 3, phys banks 1, Trfc for 2Gb part */
          mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
          mode |= MODEx_REG_DDRSDRAM_ROWBITS_14;
          mode |= MODEx_REG_DDRSDRAM_COLBITS_M4;
          mode |= MODEx_REG_DDRSDRAM_BANKBITS_3;
          mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
          mode |= MODEx_REG_DDR2SDRAM_TRFC_2Gb;
          _tprintf(_T(" Bank %lu (DDR-II SDRAM): unbuffered, row 14, col 10, bank 3, phys banks 1\n"),
            (unsigned long)i);
        } else {
          pGlob->bankTestMask &= ~(0x1U << i);
          _tprintf(_T(" +++ Bank %lu (DDR-II SDRAM): invalid size (0x%lX words), not testing\n"),
            (unsigned long)i, (unsigned long)pBankInfo[i].Size);
        }
        modes[i] = mode;
        pGlob->bBankDdr[i] = TRUE;
        break;

      default:
        assert(FALSE); /* Shouldn't get here */
        break;
      }
    }
  }
}

int
_tmain(
  int argc,
  TCHAR *argv[])
{
  Globals glob;
  int ret = 0;
  ADMXRC2_STATUS status;
  ADMXRC2_CARD_INFO cardInfo;
  ADMXRC2_SPACE_INFO spInfo;
  ADMXRC2_BANK_INFO bankInfo[MAX_BANK];
  const TCHAR* pFilename;
  TCHAR* pDesignName;
  Option options[7];
  Arguments arguments;
  unsigned int i, rep;
  unsigned int numLockFlags;
  uint64_t memorySize;
  bool_t bSetMclk, bAllow64Bit;
  uint32_t modes[MAX_BANK];
  double lclkFreq, actualLclkFreq;
  double mclkFreq, actualMclkFreq;
  unsigned long nrep;

  /***********************************************************************
  ** 1. Initialize stuff so that we know where we are if we have to     **
  **    clean up if some later step fails.                              **
  ***********************************************************************/

  glob.hCard = ADMXRC2_HANDLE_INVALID_VALUE;

  /***********************************************************************
  ** 2. Set up the data structure that defines command-line syntax.     **
  ***********************************************************************/

  options[0].key = _T("banks");
  options[0].type = Option_hex;
  options[0].def.hexVal = 0xffffU;
  options[0].meaning = _T("bitmask of banks to test");
  options[1].key = _T("card");
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("ID of card to use");
  options[2].key = _T("index");
  options[2].type = Option_uint;
  options[2].def.uintVal = 0;
  options[2].meaning = _T("index of card to use");
  options[3].key = _T("repeat");
  options[3].type = Option_uint;
  options[3].def.uintVal = 1;
  options[3].meaning = _T("number of repetitions");
  options[4].key = _T("64");
  options[4].type = Option_boolean;
  options[4].def.booleanVal = FALSE;
  options[4].meaning = _T("TRUE => use 64 bit local bus");
  options[5].key = _T("lclk");
  options[5].type = Option_float;
  options[5].def.floatVal = 33.0;
  options[5].meaning = _T("local bus clock frequency, in MHz");
  options[6].key = _T("mclk");
  options[6].type = Option_float;
  options[6].def.floatVal = 33.0;
  options[6].meaning = _T("MCLK frequency, in MHz");
  arguments.nOption = 7;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  /***********************************************************************
  ** 3. Parse the command line                                          **
  ***********************************************************************/

  parse_command_line(_T("MEMORYF example application"), _T("memoryf"), argc, argv, &arguments);

  if (options[1].specified && options[2].specified) {
    _tprintf(_T("*** Do not specify both -card and -index\n"));
    ret = -1;
    goto done;
  }

  nrep = options[3].value.uintVal;
  glob.bUse64Bit = options[4].value.booleanVal;

  /***********************************************************************
  ** 4. Open the selected card                                          **
  ***********************************************************************/

  if (!options[2].specified) {
    ADMXRC2_CARDID cardID = options[1].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &glob.hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("*** Failed to open card with ID %lu: %s\n"),
        (unsigned long)cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[2].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &glob.hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
        (unsigned long)index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  /***********************************************************************
  ** 5. Get card information for the selected card. This will be used   **
  **    later to determine default clock frequencies and other model-   **
  **    specific settings.                                              **
  ***********************************************************************/

  status = ADMXRC2_GetCardInfo(glob.hCard, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 6. Figure out model-specific settings from card type.              **
  ***********************************************************************/

  switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 33.33;
      mclkFreq = 40.0;
      numLockFlags = 2;
      break;

    case ADMXRC2_BOARD_ADMXRC_P:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 33.33;
      mclkFreq = 40.0;
      numLockFlags = 1;
      break;

    case ADMXRC2_BOARD_ADMXRC2_LITE:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 40.0;
      mclkFreq = 66.67;
      numLockFlags = 3;
      break;

    case ADMXRC2_BOARD_ADMXRC2:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 50.0;
      mclkFreq = 66.67;
      numLockFlags = 3;
      break;

    case ADMXRC2_BOARD_ADMXPL:
      bAllow64Bit = TRUE;
      bSetMclk = FALSE;
      lclkFreq = 66.67;
      mclkFreq = 2.0 * lclkFreq;
      numLockFlags = 2;
      break;

    case ADMXRC2_BOARD_ADMXP:
      bAllow64Bit = TRUE;
      bSetMclk = FALSE;
      lclkFreq = 66.67;
      mclkFreq = 2.0 * lclkFreq;
      numLockFlags = 1;
      break;

    case ADMXRC2_BOARD_ADPXPI:
      bAllow64Bit = TRUE;
      bSetMclk = FALSE;
      lclkFreq = 66.67;
      mclkFreq = 2.0 * lclkFreq;
      numLockFlags = 1;
      break;

    case ADMXRC2_BOARD_ADMXRC4LS:
    case ADMXRC2_BOARD_ADMXRC4LX:
    case ADMXRC2_BOARD_ADMXRC4SX:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 66.0;
      mclkFreq = 150.0; /* 150MHz is max 1X output frequency for DCM in low frequency mode */
      numLockFlags = 2;
      break;

    case ADMXRC2_BOARD_ADMXRC4FX:
      bAllow64Bit = TRUE;
      bSetMclk = TRUE;
      lclkFreq = 66.67;
      /* 200.0MHz MCLK => 100.0MHz in memory clock domain, 200.0MHz DDR-II SDRAM clock */
      mclkFreq = 200.0;
      numLockFlags = 2;
      break;

    case ADMXRC2_BOARD_ADPEXRC4FX:
      bAllow64Bit = TRUE;
      bSetMclk = TRUE;
      lclkFreq = 66.67;
      /* 200.0MHz MCLK => 100.0MHz in memory clock domain, 200.0MHz DDR-II SDRAM clock */
      mclkFreq = 200.0;
      numLockFlags = 4;
      break;

    case ADMXRC2_BOARD_ADMXRC5LX:
      bAllow64Bit = TRUE;
      bSetMclk = TRUE;
      lclkFreq = 66.67;
      /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
      mclkFreq = 266.67;
      numLockFlags = 2;
      break;

    case ADMXRC2_BOARD_ADMXRC5T1:
      bAllow64Bit = TRUE;
      bSetMclk = TRUE;
      lclkFreq = 66.67;
      /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
      mclkFreq = 266.67;
      numLockFlags = 2;
      break;

    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
    case ADMXRC2_BOARD_ADMXRC5T2ADV6:
    case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
      bAllow64Bit = TRUE;
      bSetMclk = TRUE;
      lclkFreq = 66.67;
      /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
      mclkFreq = 266.67;
      numLockFlags = 2;
      break;

    case ADMXRC2_BOARD_ADMXRC5TZ:
      bAllow64Bit = TRUE;
      bSetMclk = TRUE;
      lclkFreq = 66.67;
      /* 166.67MHz MCLK => 166.67MHz in memory clock domain, 166.67MHz ZBT SSRAM clock */
      mclkFreq = 166.67;
      numLockFlags = 2;
      break;

    case ADMXRC2_BOARD_ADMXRC5TDA1:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 66.67;
      /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
      mclkFreq = 266.67;
      numLockFlags = 2;
      break;

    default:
      _tprintf(_T("*** This example must be run on one of the following cards:\n\n"));
      _tprintf(_T("\tADM-XRC\n"));
      _tprintf(_T("\tADM-XRC-P\n"));
      _tprintf(_T("\tADM-XRC-II-Lite\n"));
      _tprintf(_T("\tADM-XRC-II\n"));
      _tprintf(_T("\tADM-XPL\n"));
      _tprintf(_T("\tADM-XP\n"));
      _tprintf(_T("\tADP-WRC-II\n"));
      _tprintf(_T("\tADP-DRC-II\n"));
      _tprintf(_T("\tADP-XPI\n"));
      _tprintf(_T("\tADM-XRC-4LX\n"));
      _tprintf(_T("\tADM-XRC-4SX\n"));
      _tprintf(_T("\tADM-XRC-4FX / ADM-XMC-4FX\n"));
      _tprintf(_T("\tADPE-XRC-4FX\n"));
      _tprintf(_T("\tADM-XRC-5LX\n"));
      _tprintf(_T("\tADM-XRC-5T1\n"));
      _tprintf(_T("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6\n"));
      _tprintf(_T("\tADM-XRC-5TZ\n"));
      _tprintf(_T("\tADM-XRC-5T-DA1\n"));
      _tprintf(_T("\tADM-XRC-5T2-ADV-CC1\n"));
      ret = -1;
      goto done;
  }

  if (glob.bUse64Bit && !bAllow64Bit) {
    _tprintf(_T("*** 64-bit local bus is not supported on the specified card.\n"));
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 7. Get the addresses for the FPGA registers and the memory access  **
  **    window, in our virtual address space.                           **
  ***********************************************************************/

  status = ADMXRC2_GetSpaceInfo(glob.hCard, 0, &spInfo);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to get space 0 info: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  glob.ppFpgaSpace = spInfo.VirtualBase;
  if (glob.ppFpgaSpace == NULL) {
    _tprintf(_T("*** Failed to retrieve pointer to FPGA space\n"));
    ret = -1;
    goto done;
  }
  glob.pFpgaRegs = (volatile uint32_t*)glob.ppFpgaSpace;
  glob.pMemory = (volatile uint32_t*)((uint8_t*)glob.ppFpgaSpace + 0x200000U);

  /***********************************************************************
  ** 8. Get memory bank information.                                    **
  **                                                                    **
  **    To determine the byte width of a memory bank, we first divide   **
  **    by 9. If there is no remainder, then that bank has parity bits, **
  **    and we have calculated the byte width. If there is a remainder, **
  **    then we assume the bank has no parity bits, in which case we    **
  **    should be able to calculate the byte width by dividing by 8.    **
  **                                                                    **
  **    For the ADM-XRC-II-Lite, we fudge the data a little. Each pair  **
  **    of 18-bit wide banks is combined to get a logical 32-bit wide   **
  **    bank.                                                           **
  ***********************************************************************/

  /* Get bank information */
  for (i = 0; i < cardInfo.NumRAMBank && i < MAX_BANK; i++) {
    if (cardInfo.RAMBanksFitted & (0x1UL << i)) {
      status = ADMXRC2_GetBankInfo(glob.hCard, i, &bankInfo[i]);
      if (status != ADMXRC2_SUCCESS) {
        _tprintf(_T("*** Failed to get bank %lu info: %s\n"), (unsigned long)i, ADMXRC2_GetStatusString(status));
        ret = -1;
        goto done;
      }

      if (bankInfo[i].Type & ADMXRC2_RAM_SDRAM_DDR2) {
        /*
        ** The DDR-II SSRAM memory interface for Virtex 4/5 has a user-
        ** interface that is 4 times as wide as the physical interface
        ** to the memory device.
        */
        bankInfo[i].Size >>= 1;
        bankInfo[i].Width <<= 1;
      }

      if (bankInfo[i].Type & ADMXRC2_RAM_SRAM_DDR2) {
        switch (cardInfo.BoardType) {
        case ADMXRC2_BOARD_ADMXRC5T1:
        case ADMXRC2_BOARD_ADMXRC5T2:
        case ADMXRC2_BOARD_ADMXRC5T2ADV:
        case ADMXRC2_BOARD_ADMXRC5T2ADV6:
        case ADMXRC2_BOARD_ADMXRC5TDA1:
        case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
          /*
          ** The high speed DDR-II SSRAM memory interface for Virtex 4/5
          ** has a user-interface that is 4 times as wide as the
          ** physical interface to the memory device.
          */
          bankInfo[i].Size >>= 1;
          bankInfo[i].Width <<= 1;
          break;

        default:
          assert(FALSE); /* Shouldn't get here */
          break;
        }
      }
    }
  }

  /*
  ** For the ADM-XRC-II-Lite, fix up the memory bank information to make
  ** it look like we have 32-bit wide banks (we combine two 18-bit physical
  ** banks into one logical bank).
  */
  if (cardInfo.BoardType == ADMXRC2_BOARD_ADMXRC2_LITE) {
    uint32_t banksFitted = 0x0U;

    cardInfo.NumRAMBank = 2;
    for (i = 0; i < cardInfo.NumRAMBank; i++) {
      unsigned int j = i * 2;
      bool_t bFittedA, bFittedB;
      bool_t bSame;

      bFittedA = (cardInfo.RAMBanksFitted >> j) != 0;
      bFittedB = (cardInfo.RAMBanksFitted >> (j + 1)) != 0;
      if (bFittedA && bFittedB) {
        bSame = (bankInfo[j].Width == bankInfo[j + 1].Width) && (bankInfo[j].Type == bankInfo[j + 1].Type);
        if (bSame) {
          bankInfo[i].Size = minimumOf2Macro(bankInfo[j].Size, bankInfo[j + 1].Size);
          bankInfo[i].Type = bankInfo[j].Type & bankInfo[j + 1].Type;
          bankInfo[i].Fitted = bankInfo[j].Fitted & bankInfo[j + 1].Fitted;
          banksFitted |= (0x1U << i);
        } else {
          /* This logical bank has been deemed invalid */
          _tprintf(_T("+++ Can't pair banks %lu and %lu due to different configuration\n"),
            (unsigned long)j, (unsigned long)(j + 1));
        }
      }
    }
    cardInfo.RAMBanksFitted = banksFitted;
  }

  /*
  ** Establish the bank test mask. This is the set of banks that are we
  ** have been instructed to test AND are fitted. This value may be
  ** further modified during the further stages of this application in
  ** order to remove banks that are deemed invalid.
  */
  glob.bankTestMask = options[0].value.hexVal & cardInfo.RAMBanksFitted;

  /* Process bank information */
  for (i = 0; i < cardInfo.NumRAMBank && i < MAX_BANK; i++) {
    uint32_t typeMask = 0;

    typeMask |= ADMXRC2_RAM_ZBTFT;
    typeMask |= ADMXRC2_RAM_ZBTP;
    typeMask |= ADMXRC2_RAM_SRAM_DDR2;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR2;

    if (!(glob.bankTestMask & (0x1UL << i))) {
      /* Ignore banks that we're instructed not to test */
      continue;
    }

    if (!(bankInfo[i].Type & typeMask)) {
      /* Banks that aren't fitted are invalid */
      glob.bankTestMask &= ~(0x1U << i);
      continue;
    }

    if ((bankInfo[i].Width % 9) == 0) {
      /* Bank has parity bits - don't use to calculate bank size */
      glob.bankWidth[i] = bankInfo[i].Width / 9;
    } else {
      /* Bank doesn't have parity bits */
      glob.bankWidth[i] = bankInfo[i].Width / 8;
    }
  }

  /***********************************************************************
  **  9. Program the frequencies for LCLK and MCLK. The defaults are    **
  **     obtained from the board type, but may be overridden on the     **
  **     command-line.                                                  **
  ***********************************************************************/

  if (options[5].specified) {
    lclkFreq = options[5].value.floatVal;
  }

  if (options[6].specified) {
    mclkFreq = options[6].value.floatVal;
  }

  _tprintf(_T("Setting LCLK frequency to %.1fMHz\n"), lclkFreq);
  status = ADMXRC2_SetClockRate(glob.hCard, ADMXRC2_CLOCK_LCLK, lclkFreq * 1.0e6, &actualLclkFreq);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to set LCLK to %.1fMHz: %s\n"), lclkFreq, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  if (bSetMclk) {
    _tprintf(_T("Setting MCLK frequency to %.1fMHz\n"), mclkFreq);
    status = ADMXRC2_SetClockRate(glob.hCard, 1, mclkFreq * 1.0e6, &actualMclkFreq);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("*** Failed to set clock 1 to %.1fMHz: %s\n"),
        mclkFreq, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
    _tprintf(_T("Actual MCLK frequency %.1fMHz\n"), actualMclkFreq / 1.0e6);
  }

  /***********************************************************************
  ** 10. Configure the FPGA with the "memory" bitstream.                **
  ***********************************************************************/

  pDesignName = glob.bUse64Bit ? _T("memory64") : _T("memory");
  pFilename = get_fpga_filename_dir(&cardInfo, pDesignName, pDesignName, &arguments);
  _tprintf(_T("Configuring FPGA with '%s'...\n"), pFilename);
  status = ADMXRC2_ConfigureFromFile(glob.hCard, pFilename);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to load the bitstream '%s': %s\n"), pFilename, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 11. After configuring the FPGA, it's necessary to wait a while     **
  **     to enable any DLLs, DCMs and PLLs that depend upon LCLK to     **
  **     achieve lock. The delay is a conservative 500ms, because we    **
  **     make no assumptions about whether or not an ES (engineering    **
  **     silicon) FPGA is being used.                                   **
  ***********************************************************************/

  if (!waitForLclkDcm(&glob)) {
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 12. If running in 64-bit mode was requested on the command-line,   **
  **     configure Direct Slave space 0 for a 64-bit local bus. Also    **
  **     generate the DMA mode word for DMA using a 64-bit local bus.   **
  ***********************************************************************/

  if (glob.bUse64Bit) {
    uint32_t config = 0;

    config |= ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_WIDTH_64;
    status = ADMXRC2_SetSpaceConfig(glob.hCard, 0, config);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("*** Failed to set space configuration: %s\n"), ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  /***********************************************************************
  ** 13. Here we use the memory bank information read in step 8 to      **
  **     determine the proper values for the MODE registers.            **
  ***********************************************************************/

  scanMemoryPorts(&cardInfo, bankInfo, &glob, modes);

  /***********************************************************************
  ** 14. Establish the MODE registers, based on what was determined     **
  **     in step 13.                                                    **
  ***********************************************************************/

  setMemoryPortConfiguration(&glob, modes);

  /***********************************************************************
  ** 15. Reset the memory subsystem, and wait for DLLs, DCMs, PLL and   **
  **     IDELAYCTRLs to achieve lock, as well as allowing any self-     **
  **     training memory interfaces sufficient time to train            **
  **     themselves.                                                    **
  ***********************************************************************/

  /*
  ** Reset the memory subsystem to ensure that memory banks can
  ** train themselves etc.
  */
  _tprintf(_T("Resetting memory subsystem...\n"));
  glob.pFpgaRegs[MEMCTL_REG] = 0x1U;
  glob.pFpgaRegs[MEMCTL_REG];
  sleepMs(200);
  glob.pFpgaRegs[MEMCTL_REG] = 0x0U;
  glob.pFpgaRegs[MEMCTL_REG];
  sleepMs(300);

  /* Check for memory DLL/DCM/PLL locking and training */
  if (!checkForMemoryLock(&glob, numLockFlags)) {
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 16. Figure out total size (in bytes) of memory banks that are to   **
  **     be tested.                                                     **
  ***********************************************************************/

  memorySize = 0;
  for (i = 0; i < cardInfo.NumRAMBank && i < MAX_BANK; i++) {
    uint32_t typeMask = 0;

    typeMask |= ADMXRC2_RAM_ZBTFT;
    typeMask |= ADMXRC2_RAM_ZBTP;
    typeMask |= ADMXRC2_RAM_SRAM_DDR2;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR2;

    if (!(glob.bankTestMask & (0x1UL << i))) {
      /* Ignore banks that we're instructed not to test */
      continue;
    }

    if (!bankInfo[i].Fitted) {
      /* Banks that aren't fitted are invalid */
      glob.bankTestMask &= ~(0x1U << i);
      continue;
    }

    if (!(bankInfo[i].Type & typeMask)) {
      /* Banks that aren't fitted are invalid */
      glob.bankTestMask &= ~(0x1U << i);
      continue;
    }

    glob.bankSize[i] = glob.bankWidth[i] * bankInfo[i].Size;
    memorySize += (glob.bankSize[i] >> 10);
  }

  /* If there's nothing to do, abort! */
  if (!glob.bankTestMask) {
    _tprintf(_T("*** Nothing to test - aborting.\n"));
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 17. This is where the memory tests begin. The memory test is       **
  **     performed by the FPGA for each memory bank, in six phases:     **
  **                                                                    **
  **       0 - 55 pattern                                               **
  **       1 - AA pattern                                               **
  **       2 - Alternating 55/AA pattern                                **
  **       3 - Address of 32-bit word written to each 32-bit word       **
  **       4 - Bit-reversed address of 32-bit word written to each      **
  **           32-bit word.                                             **
  **       5 - "Random" data, generated by a linear feedback shift      **
  **           register.                                                **
  ***********************************************************************/

  _tprintf(_T("Total memory size to test : %lu kiB (0x%lX kiB)\n"),
    (unsigned long)memorySize, (unsigned long)memorySize);

  for (rep = 0; rep < nrep; rep++) {
    static const TCHAR str[] = _T("|/-\\");
    int strIndex = 0;
    uint32_t val32;
    uint16_t bstat, berror;
    uint32_t eaddr;
    uint8_t ephase;
    admxrc2_time_t start, now, nextTick;

    /* Set test start address and length for each bank */
    for (i = 0; i < 16; i++) {
      if (glob.bankTestMask & (0x1U << i)) {
        glob.pFpgaRegs[USER_REG(i)] = 0;
        glob.pFpgaRegs[USER_REG(16 + i)] = bankInfo[i].Size - 1;
        glob.pFpgaRegs[USER_REG(16 + i)];
      }
    }

    /* Kick off the tests for all selected banks */
    _tprintf(_T("Commanding tests to begin (bank mask 0x%lX)...\n"), (unsigned long)glob.bankTestMask);
    glob.pFpgaRegs[USER_REG(63)] = glob.bankTestMask;
    glob.pFpgaRegs[USER_REG(63)];
    sleepMs(1);

    start = timeGet();
    nextTick = timeAdd(start, timeMake(0, 100000, TRUE));
    while (1) {
      now = timeGet();
      if (timeAfter(now, nextTick)) {
        val32 = glob.pFpgaRegs[USER_REG(63)];
        bstat = (uint16_t)val32;
        berror = (uint16_t)(val32 >> 16U);
        if ((bstat & glob.bankTestMask) == (0xffffU & glob.bankTestMask)) {
          break;
        }

        _tprintf(_T("%c Waiting for test completion (BSTAT=0x%08lX BERROR=0x%04lX)...\r"),
          str[strIndex], (unsigned long)bstat, (unsigned long)berror);
        strIndex = (strIndex + 1) & 0x3U;
        nextTick = timeAdd(nextTick, timeMake(0, 100000, TRUE));
      }

#if HOST_READ_ACTIVITY
      /*
      ** Read memory banks while we wait for completion, to see if we can provoke
      ** an error of any kind. Note that this will slow down the memory test
      ** considerably, since the local bus interface is given priority access to
      ** the memory.
      */
      for (i = 0; i < 16; i++) {
        if (glob.bankTestMask & (0x1U << i)) {
          selectBank(&glob, i);
          val32 = glob.pMemory[0];
        }
      }
#endif
    }
    now = timeGet();
    _tprintf(_T("\nDuration %lu ms\n"), (unsigned long)(timeToDouble(timeSubtract(now, start)) / 1000.0));

    val32 = glob.pFpgaRegs[USER_REG(63)];
    bstat = (uint16_t)val32;
    berror = (uint16_t)(val32 >> 16U);
    _tprintf(_T("All selected banks done - BSTAT=0x%04lX BERROR=0x%04lX\n"),
      (unsigned long)bstat, (unsigned long)berror);

    for (i = 0; i < 16; i++) {
      if (glob.bankTestMask & (0x1U << i)) {
        if (berror & (0x1U << i)) {
          eaddr = glob.pFpgaRegs[USER_REG(32 + i)];
          val32 = glob.pFpgaRegs[USER_REG(48 + (i >> 2))];
          ephase = (uint8_t)(val32 >> (i << 3));
          _tprintf(_T("*** Bank %lu failed - EADDR[%lu]=0x%08lX EPHASE[%lu]=%02lX\n"),
            (unsigned long)i,
            (unsigned long)i, (unsigned long)val32,
            (unsigned long)i, (unsigned long)ephase);
        } else {
          _tprintf(_T("Bank %lu OK\n"), (unsigned long)i);
        }
      }
    }

    if (berror & glob.bankTestMask) {
      break;
    }
  }

  /***********************************************************************
  ** 18. Here we check that none of the sticky flags in STATUS indicate **
  **     that a DLL/DCM/PLL lost lock, or that a memory bank's          **
  **     "trained" status transitioned to zero.                         **
  ***********************************************************************/

  if (!checkStickyStatus(&glob, numLockFlags)) {
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 19. Clean up.                                                      **
  ***********************************************************************/

done:
  if (glob.hCard != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(glob.hCard);
  }

  return ret;
}
