/*
** monitor.c - Trivial example showing how to read sensor values from
**             the system monitoring hardware on Gen 3 Alpha Data
**             reconfigurable computing hardware.
**
** (C) Copyright 2010-2013 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _tscanf_s(buffer, ...) _tscanf(buffer, __VA_ARGS__) /* This is OK unless the format string contains %s */
#   define _stprintf_s _sntprintf
#   define _tcscpy_s(x, y, z) _tcscpy(x, z)
# endif

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tmain main
# define _tprintf printf
# define _tscanf_s(buffer, ...) scanf(buffer, __VA_ARGS__) /* This is OK unless the format string contains %s */
# define _stprintf_s snprintf
# define _tcscpy_s(x, y, z) strcpy(x, z)
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <admxrc2.h>
#include <admxrc2/sensor.h>

#include <sdk_common.h>
#include <platform_time.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define FORMFEED (12)

static const struct {
  ADMXRC2_BOARD_TYPE boardType;
  const TCHAR* pName;
} boards[] = {
  { ADMXRC2_BOARD_ADMXRC, _T("ADM-XRC") },
  { ADMXRC2_BOARD_ADMXRC_P, _T("ADM-XRC-P") },
  { ADMXRC2_BOARD_ADMXRC2L, _T("ADM-XRC-II-Lite") },
  { ADMXRC2_BOARD_ADMXRC2, _T("ADM-XRC-II") },
  { ADMXRC2_BOARD_ADMXP, _T("ADM-XP") },
  { ADMXRC2_BOARD_ADMXPL, _T("ADM-XPL") },
  { ADMXRC2_BOARD_ADPWRC2, _T("ADP-WRC-II") },
  { ADMXRC2_BOARD_ADPDRC2, _T("ADP-DRC-II") },
  { ADMXRC2_BOARD_ADPXPI, _T("ADP-XPI") },
  { ADMXRC2_BOARD_ADMXRC4LS, _T("ADM-XRC-4LS") },
  { ADMXRC2_BOARD_ADMXRC4LX, _T("ADM-XRC-4LX") },
  { ADMXRC2_BOARD_ADMXRC4SX, _T("ADM-XRC-4SX") },
  { ADMXRC2_BOARD_ADPEXRC4FX, _T("ADPe-XRC-4FX") },
  { ADMXRC2_BOARD_ADMXRC4FX, _T("ADM-XRC-4FX") },
  { ADMXRC2_BOARD_ADMXRC5LX, _T("ADM-XRC-5LX") },
  { ADMXRC2_BOARD_ADMXRC5T1, _T("ADM-XRC-5T1") },
  { ADMXRC2_BOARD_ADMXRC5T2, _T("ADM-XRC-5T2") },
  { ADMXRC2_BOARD_ADCPXRC4LX, _T("ADCP-XRC-4LX") },
  { ADMXRC2_BOARD_ADMAMC5A2, _T("ADM-AMC-5A2") },
  { ADMXRC2_BOARD_ADMXRC5TZ, _T("ADM-XRC-5TZ") },
  { ADMXRC2_BOARD_ADCBBP, _T("ADC-BBP") },
  { ADMXRC2_BOARD_ADMXRC5T2ADV, _T("ADM-XRC-5T2-ADV") },
  { ADMXRC2_BOARD_ADMXRC5TDA1, _T("ADM-XRC-5T-DA1") },
  { ADMXRC2_BOARD_ADMXRC5T2ADV6, _T("ADM-XRC-5T2-ADV6") },
  { ADMXRC2_BOARD_ADMXRC5T2ADVCC1, _T("ADM-XRC-5T2-ADV-CC1") },
  { ADMXRC2_BOARD_ADMPCIE6S1, _T("ADM-PCIE-6S1") },
};

static const TCHAR*
getBoardName(
  ADMXRC2_BOARD_TYPE boardType)
{
  int i;

  for (i = 0; i < ARRAY_LENGTH(boards); i++) {
    if (boards[i].boardType == boardType) {
      return boards[i].pName;
    }
  }

  return NULL;
}

static const TCHAR*
getUnit(
  ADMXRC2_UNIT_TYPE unit)
{
  switch (unit) {
  case ADMXRC2_UNIT_NONE:
    return _T("");

  case ADMXRC2_UNIT_A:
    return _T("A");

  case ADMXRC2_UNIT_V:
    return _T("V");

  case ADMXRC2_UNIT_C:
    return _T("deg C");

  case ADMXRC2_UNIT_HZ:
    return _T("Hz");

  case ADMXRC2_UNIT_RPM:
    return _T("RPM");

  case ADMXRC2_UNIT_S:
    return _T("s");

  default:
    return _T("(unknown unit)");
  }
}

int
_tmain(
  int argc,
  TCHAR *argv[])
{
  int ret = 0;
  ADMXRC2_STATUS status;
  unsigned int numRep, period;
  ADMXRC2_HANDLE hCard = ADMXRC2_HANDLE_INVALID_VALUE;
  ADMXRC2_CARD_INFOEX cardInfoEx;
  ADMXRC2_SENSOR_INFO sensorInfo;
  TCHAR boardName[80];
  Option options[4];
  Arguments arguments;
  unsigned int i, rep;

  /* Set up expected options */
  options[0].key = _T("card");
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("ID of card to use");
  options[1].key = _T("index");
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("index of card to use");
  options[2].key = _T("repeat");
  options[2].type = Option_uint;
  options[2].def.uintVal = 0;
  options[2].meaning = _T("number of repetitions (0 => infinite)");
  options[3].key = _T("period");
  options[3].type = Option_uint;
  options[3].def.uintVal = 1;
  options[3].meaning = _T("update period, in seconds");
  arguments.nOption = 4;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  parse_command_line(_T("MONITOR utility"), _T("monitor"), argc, argv, &arguments);

  if (options[0].specified && options[1].specified) {
    _tprintf(_T("Do not specify both -card and -index\n"));
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Failed to open card with ID %lu: %s\n"), (unsigned long)cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  numRep = options[2].value.uintVal;
  period = options[3].value.uintVal;
  if (period < 1U) {
    period = 1U;
  }
  if (period > 100U) {
    period = 100U;
  }

  status = ADMXRC2_GetCardInfoEx(hCard, &cardInfoEx);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  if (NULL != getBoardName(cardInfoEx.BoardType)) {
    _tcscpy_s(boardName, ARRAY_LENGTH(boardName), getBoardName(cardInfoEx.BoardType));
  } else {
    _stprintf_s(boardName, ARRAY_LENGTH(boardName), _T("unknown (0x%lX)"), (unsigned long)cardInfoEx.BoardType);
  }

  for (rep = 0; numRep == 0 || rep < numRep; rep++) {
    ADMXRC2_SENSOR_VALUE sensorValue;

    if (rep >= 1) {
      sleepMs(period * 1000);
    }

    _tprintf(_T("Board type:                               %lu (0x%lX) => %s \n"), (unsigned long)cardInfoEx.BoardType, (unsigned long)cardInfoEx.BoardType, boardName);
    _tprintf(_T("Serial number:                            %lu (0x%lX)\n"), (unsigned long)cardInfoEx.SerialNum, (unsigned long)cardInfoEx.SerialNum);
    _tprintf(_T("Number of sensors:                        %lu\n"), (unsigned long)cardInfoEx.NumSensor);

    if (0 == cardInfoEx.NumSensor) {
      _tprintf(_T("*** No sensors to monitor. Exiting.\n"));
      ret = -1;
      goto done;
    }

    for (i = 0; i < cardInfoEx.NumSensor; i++) {
      _tprintf(_T("  Sensor %2u "), i);
      status = ADMXRC2_GetSensorInfo(hCard, i, &sensorInfo);
      if (status != ADMXRC2_SUCCESS) {
        _tprintf(_T("(Failed to get info: %s)\n"), ADMXRC2_GetStatusString(status));
        continue;
      }

      status = ADMXRC2_ReadSensor(hCard, i, &sensorValue);
      if (status != ADMXRC2_SUCCESS) {
        _tprintf(_T("%28s: (Failed to read sensor: %s)\n"), sensorInfo.Description, ADMXRC2_GetStatusString(status));
        continue;
      }

      switch (sensorInfo.DataType) {
      case ADMXRC2_DATA_BOOL:
        assert(sensorInfo.Exponent == 0 && sensorInfo.Unit == ADMXRC2_UNIT_NONE);
        _tprintf(_T("%28s: %s\n"), sensorInfo.Description, sensorValue.Bool ? "TRUE" : "FALSE");
        break;

      case ADMXRC2_DATA_DOUBLE:
        _tprintf(_T("%28s: %f"), sensorInfo.Description, sensorValue.Double);
        break;

      case ADMXRC2_DATA_INT32:
        _tprintf(_T("%28s: %ld"), sensorInfo.Description, (long)sensorValue.Int32);
        break;

      case ADMXRC2_DATA_UINT32:
        _tprintf(_T("%28s: %lu"), sensorInfo.Description, (unsigned long)sensorValue.UInt32);
        break;

      default:
        /* Unknown type; can't interpret */
        _tprintf(_T("%28s: ?"), sensorInfo.Description);
        break;
      }
      if (sensorInfo.DataType != ADMXRC2_DATA_BOOL) {
        if (sensorInfo.Exponent) {
          _tprintf(_T(" * 10e%d %s\n"), sensorInfo.Exponent, getUnit(sensorInfo.Unit));
        } else {
          _tprintf(_T(" %s\n"), getUnit(sensorInfo.Unit));
        }
      }
    }

    _tprintf(_T("\n"));
  }

done:
  if (hCard != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(hCard);
    hCard = ADMXRC2_HANDLE_INVALID_VALUE;
  }

  return ret;
}
