/*
** simple.c - Trivial example showing how to implement read/write registers
**            in the FPGA.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**   o ADM-XRC-5T2-ADV-CC1
**   o ADM-PCIE-6S1
**
** (C) Copyright 2000-2014 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _tmain main
# define _T(c) c
# define _tprintf printf
# define _tscanf_s(buffer, ...) scanf(buffer, __VA_ARGS__)
# define _gettchar getchar

typedef unsigned long ULONG;

#endif

#include <admxrc2.h>

#include <sdk_common.h>

#include <platform_endian.h>

int
_tmain(
  int argc,
  TCHAR *argv[])
{
	int ret = 0;
	ADMXRC2_STATUS status;
	ADMXRC2_HANDLE card = ADMXRC2_HANDLE_INVALID_VALUE;
	ADMXRC2_CARD_INFO cardInfo;
	ADMXRC2_SPACE_INFO spInfo;
	volatile uint32_t* pFpgaSpace;
	unsigned long dataOut;
  uint32_t dataIn;
	const TCHAR* filename;
	Option options[3];
	Arguments arguments;
	bool_t bUse64bit;

	/* Set up expected options */
	options[0].key = _T("card");
	options[0].type = Option_uint;
	options[0].def.uintVal = 0;
	options[0].meaning = _T("ID of card to use");
	options[1].key = _T("index");
	options[1].type = Option_uint;
	options[1].def.uintVal = 0;
	options[1].meaning = _T("index of card to use");
	options[2].key = _T("64");
	options[2].type = Option_boolean;
	options[2].def.booleanVal = FALSE;
	options[2].meaning = _T("TRUE => use 64 bit local bus");
	arguments.nOption = 3;
	arguments.option = options;
	arguments.nParamType = 0;
	arguments.minNParam = 0;

	/* Parse command line */
	parse_command_line(_T("SIMPLE example application"), _T("simple"), argc, argv, &arguments);

	if (options[0].specified && options[1].specified) {
		_tprintf(_T("Do not specify both -card and -index\n"));
		ret = -1;
		goto done;
	}

	if (!options[1].specified) {
		ADMXRC2_CARDID cardID = options[0].value.uintVal;

		status = ADMXRC2_OpenCard(cardID, &card);
		if (status != ADMXRC2_SUCCESS) {
			_tprintf(_T("Failed to open card with ID %lu: %s\n"),	(unsigned long)cardID, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	} else {
		unsigned int index = options[1].value.uintVal;

		status = ADMXRC2_OpenCardByIndex(index, &card);
		if (status != ADMXRC2_SUCCESS) {
			_tprintf(_T("Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	}

	bUse64bit = options[2].value.booleanVal;

	status = ADMXRC2_GetCardInfo(card, &cardInfo);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to get card info: %s\n"), ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}

	switch (cardInfo.BoardType) {
	case ADMXRC2_BOARD_ADMXRC:
	case ADMXRC2_BOARD_ADMXRC_P:
	case ADMXRC2_BOARD_ADMXRC2_LITE:
	case ADMXRC2_BOARD_ADMXRC2:
	case ADMXRC2_BOARD_ADPWRC2:
	case ADMXRC2_BOARD_ADPDRC2:
	case ADMXRC2_BOARD_ADMXRC4LS:
	case ADMXRC2_BOARD_ADMXRC4LX:
	case ADMXRC2_BOARD_ADMXRC4SX:
  case ADMXRC2_BOARD_ADMXRC5TZ:
	case ADMXRC2_BOARD_ADMXRC5TDA1:
		if (bUse64bit) {
			_tprintf(_T("*** The specified card cannot operate in 64 bit local bus mode\n"));
			ret = -1;
			goto done;
		}
		break;

	case ADMXRC2_BOARD_ADMPCIE6S1:
		if (bUse64bit) {
			_tprintf(_T("*** The selected card cannot be run in 64-bit mode.\n"));
			ret = -1;
			goto done;
		}
		break;

	case ADMXRC2_BOARD_ADMXPL:
	case ADMXRC2_BOARD_ADMXP:
	case ADMXRC2_BOARD_ADPXPI:
	case ADMXRC2_BOARD_ADMXRC4FX:
	case ADMXRC2_BOARD_ADPEXRC4FX:
	case ADMXRC2_BOARD_ADMXRC5LX:
	case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
  case ADMXRC2_BOARD_ADMXRC5T2ADV6:
  case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
    break;

	default:
		_tprintf(_T("*** This example must be run on one of the following cards:\n\n"));
		_tprintf(_T("\tADM-XRC\n"));
		_tprintf(_T("\tADM-XRC-P\n"));
		_tprintf(_T("\tADM-XRC-II-Lite\n"));
		_tprintf(_T("\tADM-XRC-II\n"));
		_tprintf(_T("\tADM-XPL\n"));
		_tprintf(_T("\tADM-XP\n"));
		_tprintf(_T("\tADP-WRC-II\n"));
		_tprintf(_T("\tADP-DRC-II\n"));
		_tprintf(_T("\tADP-XPI\n"));
		_tprintf(_T("\tADM-XRC-4LX\n"));
		_tprintf(_T("\tADM-XRC-4SX\n"));
		_tprintf(_T("\tADM-XRC-4FX / ADM-XMC-4FX\n"));
		_tprintf(_T("\tADPE-XRC-4FX\n"));
		_tprintf(_T("\tADM-XRC-5LX\n"));
		_tprintf(_T("\tADM-XRC-5T1\n"));
		_tprintf(_T("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6\n"));
		_tprintf(_T("\tADM-XRC-5TZ\n"));
		_tprintf(_T("\tADM-XRC-5T-DA1\n"));
		_tprintf(_T("\tADM-XRC-5T2-ADV-CC1\n"));
		_tprintf(_T("\tADM-PCIE-6S1\n"));
		ret = -1;
		goto done;
	}

	/* Get the address of FPGA space */
	status = ADMXRC2_GetSpaceInfo(card, 0, &spInfo);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to get space 0 info: %s\n"), ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
	pFpgaSpace = (volatile uint32_t*) spInfo.VirtualBase;
	if (pFpgaSpace == NULL) {
		_tprintf(_T("*** Failed to retrieve pointer to FPGA space\n"));
		ret = -1;
		goto done;
	}

  if (cardInfo.NumClock > 0) {
	  /* Set local bus clock to nominal frequency */
	  status = ADMXRC2_SetClockRate(card, ADMXRC2_CLOCK_LCLK, 33.0e6, NULL);
	  if (status != ADMXRC2_SUCCESS) {
		  _tprintf(_T("Failed to set LCLK to %.1fMHz: %s\n"), 33.0f, ADMXRC2_GetStatusString(status));
		  ret = -1;
		  goto done;
	  }
  }
	
	if (bUse64bit) {
		filename = get_fpga_filename(&cardInfo, _T("simple64"), &arguments);
	} else {
		filename = get_fpga_filename(&cardInfo, _T("simple"), &arguments);
	}
	status = ADMXRC2_ConfigureFromFile(card, filename);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to load the bitstream '%s': %s\n"), filename, ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}

	if (bUse64bit) {
		uint32_t	config;

		config = ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_WIDTH_64;
		status = ADMXRC2_SetSpaceConfig(card, 0, config);
		if (status != ADMXRC2_SUCCESS) {
			_tprintf(_T("*** Failed to set space configuration: %s\n"), ADMXRC2_GetStatusString(status));
			return -1;
		}
	}

	_tprintf(_T("====================\n"));
	_tprintf(_T("Enter values for I/O\n"));
	_tprintf(_T("(use 55AA to exit)\n"));
	_tprintf(_T("====================\n"));

	while (1) {
		_tscanf_s(_T("%lx"), &dataOut);
		pFpgaSpace[0] = cpu_to_le32((uint32_t)dataOut);
		dataIn = pFpgaSpace[0];
		_tprintf(_T("Out = 0x%08lX, In = 0x%08lX\n"), (unsigned long)dataOut, (unsigned long)le32_to_cpu(dataIn));
		if (dataOut == 0x55AAU) {
			break;
		}
	}

done:
	if (card != ADMXRC2_HANDLE_INVALID_VALUE) {
		ADMXRC2_CloseCard(card);
	}
	
	return ret;
}
